__copyright__ = "Copyright (c) 2018-2025 Alex Laird"
__license__ = "MIT"

import os
import time
import traceback
import unittest
import uuid
from http import HTTPStatus
from unittest import mock
from urllib.parse import urlparse
from urllib.request import urlopen

import yaml

from pyngrok import __version__, installer, ngrok, process
from pyngrok.conf import PyngrokConfig
from pyngrok.exception import PyngrokError, PyngrokNgrokError, PyngrokNgrokHTTPError, PyngrokNgrokURLError, \
    PyngrokSecurityError
from scripts.create_test_resources import create_test_resources, generate_name_for_subdomain
from tests.testcase import NgrokTestCase


class TestNgrok(NgrokTestCase):
    @classmethod
    def setUpClass(cls):
        if os.environ.get("NGROK_API_KEY"):
            testcase_config_dir = os.path.normpath(
                os.path.join(os.path.abspath(os.path.dirname(__file__)), ".testcase-ngrok"))
            testcase_config_path = os.path.join(testcase_config_dir, "config.yml")
            testcase_ngrok_path = os.path.join(testcase_config_dir, installer.get_ngrok_bin())
            cls.testcase_pyngrok_config = PyngrokConfig(ngrok_path=testcase_ngrok_path,
                                                        config_path=testcase_config_path)
            cls.given_ngrok_installed(cls.testcase_pyngrok_config)

            # NGROK_HOSTNAME is set when init_test_resources.py is done provisioning test resources, so if it
            # hasn't been set, we need to do that now. When running tests on CI, using the init script can protect
            # against rate limiting, as this allows API resources to be shared across the build matrix.
            if not os.environ.get("NGROK_HOSTNAME"):
                create_test_resources(True)

                cls.reserved_domain_id = os.environ["NGROK_DOMAIN_ID"]

            cls.reserved_domain = os.environ["NGROK_DOMAIN"]

    @classmethod
    def tearDownClass(cls):
        # NGROK_HOSTNAME is set when init_test_resources.py is done provisioning test resources, in which case
        # prune_test_resources.py should also be called to clean up test resources after all tests complete. Otherwise,
        # this testcase set up the resources, so it should also tear them down.
        if os.environ.get("NGROK_API_KEY") and not os.environ.get("NGROK_HOSTNAME"):
            try:
                ngrok.api("reserved-domains", "delete", cls.reserved_domain_id,
                          pyngrok_config=cls.testcase_pyngrok_config)
            except Exception:
                print(traceback.format_exc())
                print("--> An error occurred while cleaning up test resources. Run scripts/prune_test_resources.py to "
                      "finish.")

    @mock.patch("subprocess.call")
    def test_run(self, mock_call):
        # WHEN
        ngrok.run()

        # THEN
        self.assertTrue(mock_call.called)

    @mock.patch("subprocess.call")
    def test_main(self, mock_call):
        # WHEN
        ngrok.main()

        # THEN
        self.assertTrue(mock_call.called)

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_connect_v2(self):
        # GIVEN
        self.assertEqual(len(process._current_processes.keys()), 0)
        self.assertEqual(len(ngrok._current_tunnels.keys()), 0)

        # WHEN
        ngrok_tunnel = ngrok.connect("5000", pyngrok_config=self.pyngrok_config_v2)
        current_process = ngrok.get_ngrok_process(pyngrok_config=self.pyngrok_config_v2)

        # THEN
        self.assertEqual(len(ngrok._current_tunnels.keys()), 1)
        self.assertIsNotNone(current_process)
        self.assertIsNone(current_process.proc.poll())
        self.assertTrue(current_process._monitor_thread.is_alive())
        self.assertIsNone(ngrok_tunnel.id)
        self.assertTrue(ngrok_tunnel.name.startswith("http-5000-"))
        self.assertEqual("http", ngrok_tunnel.proto)
        self.assertEqual("http://localhost:5000", ngrok_tunnel.config["addr"])
        self.assertIsNotNone(ngrok_tunnel.public_url)
        self.assertIsNotNone(process.get_process(self.pyngrok_config_v2))
        self.assertIn('http://', ngrok_tunnel.public_url)
        self.assertEqual(len(process._current_processes.keys()), 1)

        # WHEN
        ngrok.kill(self.pyngrok_config_v2)
        ngrok_version, pyngrok_version = ngrok.get_version(pyngrok_config=self.pyngrok_config_v2)

        # THEN
        self.assertTrue(ngrok_version.startswith("2"))

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_connect_v3(self):
        # GIVEN
        self.assertEqual(len(process._current_processes.keys()), 0)
        self.assertEqual(len(ngrok._current_tunnels.keys()), 0)

        # WHEN
        ngrok_tunnel = ngrok.connect("5000", pyngrok_config=self.pyngrok_config_v3)
        current_process = ngrok.get_ngrok_process(self.pyngrok_config_v3)

        # THEN
        self.assertEqual(len(ngrok._current_tunnels.keys()), 1)
        self.assertIsNotNone(current_process)
        self.assertIsNone(current_process.proc.poll())
        self.assertTrue(current_process._monitor_thread.is_alive())
        self.assertIsNotNone(ngrok_tunnel.id)
        self.assertTrue(ngrok_tunnel.name.startswith("http-5000-"))
        self.assertEqual("https", ngrok_tunnel.proto)
        self.assertEqual("http://localhost:5000", ngrok_tunnel.config["addr"])
        self.assertIsNotNone(ngrok_tunnel.public_url)
        self.assertIsNotNone(process.get_process(self.pyngrok_config_v3))
        self.assertIn('https://', ngrok_tunnel.public_url)
        self.assertEqual(len(process._current_processes.keys()), 1)

        # WHEN
        ngrok.kill(self.pyngrok_config_v3)
        ngrok_version, pyngrok_version = ngrok.get_version(pyngrok_config=self.pyngrok_config_v3)

        # THEN
        self.assertTrue(ngrok_version.startswith("3"))

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    @unittest.skipIf(not os.environ.get("NGROK_API_KEY"), "NGROK_API_KEY environment variable not set")
    def test_connect_tls(self):
        # GIVEN
        self.assertEqual(len(process._current_processes.keys()), 0)
        self.assertEqual(len(ngrok._current_tunnels.keys()), 0)

        # WHEN
        ngrok_tunnel = ngrok.connect("443", proto="tls", domain=self.reserved_domain,
                                     terminate_at="upstream",
                                     pooling_enabled=True, pyngrok_config=self.pyngrok_config_v3)
        current_process = ngrok.get_ngrok_process(self.pyngrok_config_v3)

        # THEN
        self.assertEqual(len(ngrok._current_tunnels.keys()), 1)
        self.assertIsNotNone(current_process)
        self.assertIsNone(current_process.proc.poll())
        self.assertTrue(current_process._monitor_thread.is_alive())
        self.assertIsNotNone(ngrok_tunnel.id)
        self.assertTrue(ngrok_tunnel.name.startswith("tls-443-"))
        self.assertEqual("tls", ngrok_tunnel.proto)
        self.assertEqual("tls://localhost:443", ngrok_tunnel.config["addr"])
        self.assertIsNotNone(ngrok_tunnel.public_url)
        self.assertIsNotNone(process.get_process(self.pyngrok_config_v3))
        self.assertIn('tls://', ngrok_tunnel.public_url)
        self.assertEqual(len(process._current_processes.keys()), 1)

        # WHEN
        ngrok.kill(self.pyngrok_config_v3)
        ngrok_version, pyngrok_version = ngrok.get_version(pyngrok_config=self.pyngrok_config_v3)

        # THEN
        self.assertTrue(ngrok_version.startswith("3"))

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_connect_name(self):
        # WHEN
        ngrok_tunnel = ngrok.connect(name="my-tunnel", pyngrok_config=self.pyngrok_config_v3)

        # THEN
        self.assertEqual(ngrok_tunnel.name, "my-tunnel")
        self.assertEqual("https", ngrok_tunnel.proto)
        self.assertEqual("http://localhost:80", ngrok_tunnel.config["addr"])

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_get_tunnels(self):
        # GIVEN
        url = ngrok.connect(pyngrok_config=self.pyngrok_config_v3).public_url
        time.sleep(1)
        self.assertEqual(len(ngrok._current_tunnels.keys()), 1)

        # WHEN
        tunnels = ngrok.get_tunnels(self.pyngrok_config_v3)

        # THEN
        self.assertEqual(len(ngrok._current_tunnels.keys()), 1)
        self.assertEqual(len(tunnels), 1)
        self.assertEqual(tunnels[0].proto, "https")
        self.assertEqual(tunnels[0].public_url, url)
        self.assertEqual(tunnels[0].config["addr"], "http://localhost:80")

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_bind_tls_both_v2(self):
        # WHEN
        url = ngrok.connect(bind_tls="both", pyngrok_config=self.pyngrok_config_v2).public_url
        num_tunnels = len(ngrok.get_tunnels(self.pyngrok_config_v2))

        # THEN
        self.assertTrue(url.startswith("http://"))
        self.assertEqual(num_tunnels, 2)

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_bind_tls_https_only_v2(self):
        # WHEN
        url = ngrok.connect(bind_tls=True, pyngrok_config=self.pyngrok_config_v2).public_url
        num_tunnels = len(ngrok.get_tunnels(self.pyngrok_config_v2))

        # THEN
        self.assertTrue(url.startswith("https://"))
        self.assertEqual(num_tunnels, 1)

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_bind_tls_http_only_v2(self):
        # WHEN
        url = ngrok.connect(bind_tls=False, pyngrok_config=self.pyngrok_config_v2).public_url
        num_tunnels = len(ngrok.get_tunnels(self.pyngrok_config_v2))

        # THEN
        self.assertTrue(url.startswith("http://"))
        self.assertEqual(num_tunnels, 1)

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_schemes_http_v3(self):
        # WHEN
        url = ngrok.connect(schemes=["http"], pyngrok_config=self.pyngrok_config_v3).public_url
        num_tunnels = len(ngrok.get_tunnels(self.pyngrok_config_v3))

        # THEN
        self.assertTrue(url.startswith("http://"))
        self.assertEqual(num_tunnels, 1)

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_schemes_https_v3(self):
        # WHEN
        url = ngrok.connect(schemes=["https"], pyngrok_config=self.pyngrok_config_v3).public_url
        num_tunnels = len(ngrok.get_tunnels(self.pyngrok_config_v3))

        # THEN
        self.assertTrue(url.startswith("https://"))
        self.assertEqual(num_tunnels, 1)

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_schemes_http_https_v3(self):
        # WHEN
        url = ngrok.connect(schemes=["http", "https"], pyngrok_config=self.pyngrok_config_v3).public_url
        num_tunnels = len(ngrok.get_tunnels(self.pyngrok_config_v3))

        # THEN
        self.assertTrue(url.startswith("https://"))
        self.assertEqual(num_tunnels, 2)

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_bind_tls_upgraded_to_schemes_v3(self):
        # WHEN
        public_url1 = ngrok.connect(bind_tls="both", pyngrok_config=self.pyngrok_config_v3).public_url
        num_tunnels1 = len(ngrok.get_tunnels(self.pyngrok_config_v3))
        public_url2 = ngrok.connect(bind_tls=True, pyngrok_config=self.pyngrok_config_v3).public_url
        num_tunnels2 = len(ngrok.get_tunnels(self.pyngrok_config_v3))
        public_url3 = ngrok.connect(bind_tls=False, pyngrok_config=self.pyngrok_config_v3).public_url
        num_tunnels3 = len(ngrok.get_tunnels(self.pyngrok_config_v3))

        # THEN
        self.assertTrue(public_url1.startswith("https://"))
        self.assertEqual(num_tunnels1, 2)
        self.assertTrue(public_url2.startswith("https://"))
        self.assertEqual(num_tunnels2, 3)
        self.assertTrue(public_url3.startswith("http://"))
        self.assertEqual(num_tunnels3, 4)

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_disconnect_v2(self):
        # GIVEN
        url = ngrok.connect(pyngrok_config=self.pyngrok_config_v2).public_url
        time.sleep(1)
        tunnels = ngrok.get_tunnels(self.pyngrok_config_v2)
        # Two tunnels, as one each was created for "http" and "https"
        self.assertEqual(len(ngrok._current_tunnels.keys()), 2)
        self.assertEqual(len(tunnels), 2)

        # WHEN
        ngrok.disconnect(url, pyngrok_config=self.pyngrok_config_v2)
        self.assertEqual(len(ngrok._current_tunnels.keys()), 1)
        time.sleep(1)
        tunnels = ngrok.get_tunnels(self.pyngrok_config_v2)

        # THEN
        # There is still one tunnel left, as we only disconnected the http tunnel
        self.assertEqual(len(ngrok._current_tunnels.keys()), 1)
        self.assertEqual(len(tunnels), 1)

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_disconnect_v3(self):
        # GIVEN
        url = ngrok.connect(pyngrok_config=self.pyngrok_config_v3).public_url
        time.sleep(1)
        tunnels = ngrok.get_tunnels(self.pyngrok_config_v3)
        self.assertEqual(len(ngrok._current_tunnels.keys()), 1)
        self.assertEqual(len(tunnels), 1)

        # WHEN
        ngrok.disconnect(url, self.pyngrok_config_v3)
        self.assertEqual(len(ngrok._current_tunnels.keys()), 0)
        time.sleep(1)
        tunnels = ngrok.get_tunnels(self.pyngrok_config_v3)

        # THEN
        self.assertEqual(len(ngrok._current_tunnels.keys()), 0)
        self.assertEqual(len(tunnels), 0)

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_kill(self):
        # GIVEN
        ngrok.connect("5000", pyngrok_config=self.pyngrok_config_v3)
        time.sleep(1)
        ngrok_process = process.get_process(self.pyngrok_config_v3)
        monitor_thread = ngrok_process._monitor_thread
        self.assertEqual(len(ngrok._current_tunnels.keys()), 1)

        # WHEN
        ngrok.kill(self.pyngrok_config_v3)
        time.sleep(1)

        # THEN
        self.assertEqual(len(ngrok._current_tunnels.keys()), 0)
        self.assertIsNotNone(ngrok_process.proc.poll())
        self.assertFalse(monitor_thread.is_alive())
        self.assertEqual(len(process._current_processes.keys()), 0)
        self.assert_no_zombies()

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_api_get_request_success(self):
        # GIVEN
        current_process = ngrok.get_ngrok_process(pyngrok_config=self.pyngrok_config_v3)
        ngrok_tunnel = ngrok.connect(pyngrok_config=self.pyngrok_config_v3)
        time.sleep(1)

        # WHEN
        response = ngrok.api_request(f"{current_process.api_url}{ngrok_tunnel.uri}", "GET")

        # THEN
        self.assertEqual(ngrok_tunnel.name, response["name"])
        self.assertEqual(ngrok_tunnel.public_url, response["public_url"])

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_api_request_query_params(self):
        # GIVEN
        tunnel_name = "tunnel (1)"
        current_process = ngrok.get_ngrok_process(pyngrok_config=self.pyngrok_config_v3)
        public_url = ngrok.connect(urlparse(current_process.api_url).port, name=tunnel_name,
                                   pyngrok_config=self.pyngrok_config_v3).public_url
        time.sleep(1)

        urlopen(f"{public_url}/api/status").read()
        time.sleep(3)

        # WHEN
        response1 = ngrok.api_request(f"{current_process.api_url}/api/requests/http", "GET")
        response2 = ngrok.api_request(f"{current_process.api_url}/api/requests/http", "GET",
                                      params={"tunnel_name": f"{tunnel_name}"})
        response3 = ngrok.api_request(f"{current_process.api_url}/api/requests/http", "GET",
                                      params={"tunnel_name": "unknown-tunnel"})

        # THEN
        self.assertEqual(len(response1["requests"]), 1)
        self.assertEqual(len(response2["requests"]), 1)
        self.assertEqual(0, len(response3["requests"]))

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_api_request_delete_data_updated(self):
        # GIVEN
        current_process = ngrok.get_ngrok_process(pyngrok_config=self.pyngrok_config_v3)
        ngrok.connect(pyngrok_config=self.pyngrok_config_v3)
        time.sleep(1)
        tunnels = ngrok.get_tunnels(self.pyngrok_config_v3)
        self.assertEqual(len(tunnels), 1)

        # WHEN
        response = ngrok.api_request(f"{current_process.api_url}{tunnels[0].uri}",
                                     "DELETE")

        # THEN
        self.assertEqual(response, {})
        tunnels = ngrok.get_tunnels(self.pyngrok_config_v3)
        self.assertEqual(len(tunnels), 0)

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_api_request_fails(self):
        # GIVEN
        current_process = ngrok.get_ngrok_process(pyngrok_config=self.pyngrok_config_v3)
        bad_data = {
            "name": str(uuid.uuid4()),
            "addr": "8080",
            "proto": "invalid-proto"
        }

        # WHEN
        with self.assertRaises(PyngrokNgrokHTTPError) as cm:
            ngrok.api_request(f"{current_process.api_url}/api/tunnels", "POST", data=bad_data)

        # THEN
        self.assertEqual(HTTPStatus.BAD_REQUEST, cm.exception.status_code)
        self.assertIn("invalid tunnel configuration", str(cm.exception))
        self.assertIn("protocol name", str(cm.exception))

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_api_request_timeout(self):
        # GIVEN
        current_process = ngrok.get_ngrok_process(pyngrok_config=self.pyngrok_config_v3)
        ngrok_tunnel = ngrok.connect(pyngrok_config=self.pyngrok_config_v3)
        time.sleep(1)

        # WHEN
        with self.assertRaises(PyngrokNgrokURLError) as cm:
            ngrok.api_request(f"{current_process.api_url}{ngrok_tunnel.uri}", "DELETE",
                              timeout=0.0001)

        # THEN
        self.assertIn("timed out", cm.exception.reason)

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_regional_tcp_v2(self):
        # GIVEN
        self.assertEqual(len(process._current_processes.keys()), 0)
        subdomain = generate_name_for_subdomain("pyngrok-temp")
        pyngrok_config = self.copy_with_updates(self.pyngrok_config_v2,
                                                region="au")

        # WHEN
        ngrok_tunnel = ngrok.connect("5000", "tcp", subdomain=subdomain, pyngrok_config=pyngrok_config)
        current_process = ngrok.get_ngrok_process(pyngrok_config)

        # THEN
        self.assertIsNotNone(current_process)
        self.assertIsNone(current_process.proc.poll())
        self.assertIsNotNone(ngrok_tunnel.public_url)
        self.assertIsNotNone(process.get_process(pyngrok_config))
        self.assertEqual("localhost:5000", ngrok_tunnel.config["addr"])
        self.assertIn("tcp://", ngrok_tunnel.public_url)
        self.assertIn(".au.", ngrok_tunnel.public_url)
        self.assertEqual(len(process._current_processes.keys()), 1)

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_regional_tcp_v3(self):
        # GIVEN
        self.assertEqual(len(process._current_processes.keys()), 0)
        pyngrok_config = self.copy_with_updates(self.pyngrok_config_v3,
                                                region="au")

        # WHEN
        ngrok_tunnel = ngrok.connect("5000", "tcp", pyngrok_config=pyngrok_config)
        current_process = ngrok.get_ngrok_process(pyngrok_config)

        # THEN
        self.assertIsNotNone(current_process)
        self.assertIsNone(current_process.proc.poll())
        self.assertIsNotNone(ngrok_tunnel.public_url)
        self.assertIsNotNone(process.get_process(pyngrok_config))
        self.assertEqual("localhost:5000", ngrok_tunnel.config["addr"])
        self.assertIn("tcp://", ngrok_tunnel.public_url)
        self.assertIn(".au.", ngrok_tunnel.public_url)
        self.assertEqual(len(process._current_processes.keys()), 1)

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_auth_v2(self):
        # GIVEN
        self.assertEqual(len(process._current_processes.keys()), 0)

        # WHEN
        ngrok_tunnel = ngrok.connect("5000", auth="username:password", pyngrok_config=self.pyngrok_config_v2)
        current_process = ngrok.get_ngrok_process(self.pyngrok_config_v2)

        # THEN
        self.assertIsNotNone(current_process)
        self.assertIsNone(current_process.proc.poll())
        self.assertIsNotNone(ngrok_tunnel.public_url)
        self.assertIsNotNone(process.get_process(self.pyngrok_config_v2))
        self.assertEqual(len(process._current_processes.keys()), 1)

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_auth_v3(self):
        # GIVEN
        self.assertEqual(len(process._current_processes.keys()), 0)

        # WHEN
        ngrok_tunnel1 = ngrok.connect(auth="username:password", pyngrok_config=self.pyngrok_config_v3)
        ngrok_tunnel2 = ngrok.connect("5000", auth=["username:password"], pyngrok_config=self.pyngrok_config_v3)
        ngrok_tunnel3 = ngrok.connect("5001", basic_auth=["username:password"], pyngrok_config=self.pyngrok_config_v3)
        current_process = ngrok.get_ngrok_process(self.pyngrok_config_v3)

        # THEN
        self.assertIsNotNone(current_process)
        self.assertIsNone(current_process.proc.poll())
        self.assertIsNotNone(ngrok_tunnel1.public_url)
        self.assertIsNotNone(ngrok_tunnel2.public_url)
        self.assertIsNotNone(ngrok_tunnel3.public_url)
        self.assertIsNotNone(process.get_process(self.pyngrok_config_v3))
        self.assertEqual(len(process._current_processes.keys()), 1)

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_regional_subdomain(self):
        # GIVEN
        self.assertEqual(len(process._current_processes.keys()), 0)
        subdomain = generate_name_for_subdomain("pyngrok-temp")
        pyngrok_config = self.copy_with_updates(self.pyngrok_config_v3,
                                                region="au")

        # WHEN
        url = ngrok.connect("5000", subdomain=subdomain, pyngrok_config=pyngrok_config).public_url
        current_process = ngrok.get_ngrok_process(pyngrok_config)

        # THEN
        self.assertIsNotNone(current_process)
        self.assertIsNone(current_process.proc.poll())
        self.assertIsNotNone(url)
        self.assertIsNotNone(process.get_process(pyngrok_config))
        self.assertIn("https://", url)
        self.assertIn(".au.", url)
        self.assertIn(subdomain, url)
        self.assertEqual(len(process._current_processes.keys()), 1)

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_connect_fileserver(self):
        # GIVEN
        self.assertEqual(len(process._current_processes.keys()), 0)

        # WHEN
        ngrok_tunnel = ngrok.connect("file:///", pyngrok_config=self.pyngrok_config_v3)
        current_process = ngrok.get_ngrok_process(self.pyngrok_config_v3)
        tunnels = ngrok.get_tunnels(self.pyngrok_config_v3)

        # THEN
        self.assertEqual(len(tunnels), 1)
        self.assertIsNotNone(current_process)
        self.assertIsNone(current_process.proc.poll())
        self.assertTrue(current_process._monitor_thread.is_alive())
        self.assertTrue(ngrok_tunnel.name.startswith("http-file-"))
        self.assertEqual("file:///", ngrok_tunnel.config["addr"])
        self.assertIsNotNone(ngrok_tunnel.public_url)
        self.assertIsNotNone(process.get_process(self.pyngrok_config_v3))
        self.assertIn('https://', ngrok_tunnel.public_url)
        self.assertEqual(len(process._current_processes.keys()), 1)

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_disconnect_fileserver_v2(self):
        # GIVEN
        self.assertEqual(len(process._current_processes.keys()), 0)
        url = ngrok.connect("file:///", pyngrok_config=self.pyngrok_config_v2).public_url

        # WHEN
        ngrok.disconnect(url, self.pyngrok_config_v2)
        tunnels = ngrok.get_tunnels(self.pyngrok_config_v2)

        # THEN
        # There is still one tunnel left, as we only disconnected the http tunnel
        self.assertEqual(len(tunnels), 1)

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_get_tunnel_fileserver(self):
        # GIVEN
        self.assertEqual(len(process._current_processes.keys()), 0)
        ngrok_tunnel = ngrok.connect("file:///", pyngrok_config=self.pyngrok_config_v3)
        time.sleep(1)
        api_url = ngrok.get_ngrok_process(self.pyngrok_config_v3).api_url

        # WHEN
        response = ngrok.api_request(f"{api_url}{ngrok_tunnel.uri}", "GET")

        # THEN
        self.assertEqual(ngrok_tunnel.name, response["name"])
        self.assertTrue(ngrok_tunnel.name.startswith("http-file-"))

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_ngrok_tunnel_refresh_metrics(self):
        # GIVEN
        current_process = ngrok.get_ngrok_process(pyngrok_config=self.pyngrok_config_v3)
        ngrok_tunnel = ngrok.connect(urlparse(current_process.api_url).port, pyngrok_config=self.pyngrok_config_v3)
        time.sleep(1)
        self.assertEqual(0, ngrok_tunnel.metrics.get("http").get("count"))
        self.assertEqual(ngrok_tunnel.data["metrics"].get("http").get("count"), 0)

        urlopen(f"{ngrok_tunnel.public_url}/api/status").read()
        time.sleep(3)

        # WHEN
        ngrok_tunnel.refresh_metrics()

        # THEN
        self.assertGreater(ngrok_tunnel.metrics.get("http").get("count"), 0)
        self.assertGreater(ngrok_tunnel.data["metrics"].get("http").get("count"), 0)

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_tunnel_definitions_v2(self):
        subdomain = generate_name_for_subdomain("pyngrok-temp")

        # GIVEN
        config = {
            "tunnels": {
                "http-tunnel": {
                    "proto": "http",
                    "addr": "8000",
                    "subdomain": subdomain,
                    "inspect": False,
                    "bind_tls": True
                },
                "tcp-tunnel": {
                    "proto": "tcp",
                    "addr": "22"
                }
            }
        }
        config_path = os.path.join(self.config_dir, "config_v2_2.yml")
        installer.install_default_config(config_path, config, ngrok_version="v2")
        pyngrok_config = self.copy_with_updates(self.pyngrok_config_v2,
                                                config_path=config_path)

        # WHEN
        http_tunnel = ngrok.connect(name="http-tunnel", pyngrok_config=pyngrok_config)
        ssh_tunnel = ngrok.connect(name="tcp-tunnel", pyngrok_config=pyngrok_config)

        # THEN
        self.assertEqual(http_tunnel.name, "http-tunnel-api")
        self.assertEqual(http_tunnel.config["addr"],
                         f"http://localhost:{config['tunnels']['http-tunnel']['addr']}")
        self.assertEqual(http_tunnel.proto, "https")
        self.assertEqual(http_tunnel.public_url,
                         f"https://{config['tunnels']['http-tunnel']['subdomain']}.ngrok.io")
        self.assertEqual(ssh_tunnel.name, "tcp-tunnel-api")
        self.assertEqual(ssh_tunnel.config["addr"],
                         f"localhost:{config['tunnels']['tcp-tunnel']['addr']}")
        self.assertEqual(ssh_tunnel.proto, config["tunnels"]["tcp-tunnel"]["proto"])
        self.assertTrue(ssh_tunnel.public_url.startswith("tcp://"))

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_tunnel_definitions_v3(self):
        subdomain = generate_name_for_subdomain("pyngrok-temp")

        # GIVEN
        config = {
            "tunnels": {
                "http-tunnel": {
                    "proto": "http",
                    "addr": "8000",
                    "subdomain": subdomain,
                    "circuit_breaker": 0.5,
                    "oauth": {
                        "provider": "google",
                        "allow_domains": ["pyngrok.com"],
                        "allow_emails": ["email@pyngrok.com"]
                    }
                },
                "tcp-tunnel": {
                    "proto": "tcp",
                    "addr": "22"
                }
            }
        }
        config_path = os.path.join(self.config_dir, "config_v3_2.yml")
        installer.install_default_config(config_path, config, ngrok_version="v3")
        pyngrok_config = self.copy_with_updates(self.pyngrok_config_v3,
                                                config_path=config_path)

        # WHEN
        http_tunnel = ngrok.connect(name="http-tunnel", pyngrok_config=pyngrok_config)
        ssh_tunnel = ngrok.connect(name="tcp-tunnel", pyngrok_config=pyngrok_config)
        time.sleep(1)
        response_body = urlopen(http_tunnel.public_url).read().decode()

        # THEN
        self.assertEqual(http_tunnel.name, "http-tunnel-api")
        self.assertEqual(http_tunnel.config["addr"],
                         f"http://localhost:{config['tunnels']['http-tunnel']['addr']}")
        self.assertTrue(http_tunnel.config["addr"].startswith("http://"))
        self.assertEqual(http_tunnel.public_url,
                         f"https://{config['tunnels']['http-tunnel']['subdomain']}.ngrok.io")
        self.assertIn("Sign in - Google Accounts", response_body)
        self.assertEqual(ssh_tunnel.name, "tcp-tunnel-api")
        self.assertEqual(ssh_tunnel.config["addr"],
                         f"localhost:{config['tunnels']['tcp-tunnel']['addr']}")
        self.assertEqual(ssh_tunnel.proto, config["tunnels"]["tcp-tunnel"]["proto"])
        self.assertTrue(ssh_tunnel.public_url.startswith("tcp://"))

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    @unittest.skipIf(not os.environ.get("NGROK_API_KEY"), "NGROK_API_KEY environment variable not set")
    def test_tunnel_definitions_tls(self):
        # GIVEN
        config = {
            "tunnels": {
                "tls-tunnel": {
                    "proto": "tls",
                    "addr": "80",
                    "domain": self.reserved_domain,
                    "terminate_at": "upstream",
                    "pooling_enabled": True
                }
            }
        }
        config_path = os.path.join(self.config_dir, "config_v3_2.yml")
        installer.install_default_config(config_path, config, ngrok_version="v3")
        pyngrok_config = self.copy_with_updates(self.pyngrok_config_v3,
                                                config_path=config_path)

        # WHEN
        tls_tunnel = ngrok.connect(name="tls-tunnel", pyngrok_config=pyngrok_config)

        # THEN
        self.assertTrue(tls_tunnel.name.startswith("tls-"))
        self.assertEqual(tls_tunnel.config["addr"],
                         f"localhost:{config['tunnels']['tls-tunnel']['addr']}")
        self.assertTrue(tls_tunnel.config["addr"], "localhost:80")
        self.assertEqual(tls_tunnel.name, "tls-tunnel-api")
        self.assertEqual(tls_tunnel.config["addr"],
                         f"localhost:{config['tunnels']['tls-tunnel']['addr']}")
        self.assertEqual(tls_tunnel.proto, config["tunnels"]["tls-tunnel"]["proto"])
        self.assertTrue(tls_tunnel.public_url, "tls://{domain}".format(domain=self.reserved_domain))

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_ngrok_http_internal_endpoint_pooling(self):
        # WHEN
        internal_http_endpoint = ngrok.connect(proto="http", addr="80", domain="pyngrok.internal",
                                               pooling_enabled=True,
                                               pyngrok_config=self.pyngrok_config_v3)
        tunnels = ngrok.get_tunnels(pyngrok_config=self.pyngrok_config_v3)

        # THEN
        self.assertEqual(internal_http_endpoint.config["addr"], "http://localhost:80")
        self.assertTrue(internal_http_endpoint.config["addr"].startswith("http://"))
        self.assertEqual(internal_http_endpoint.proto, "https")
        self.assertEqual(internal_http_endpoint.public_url, "https://pyngrok.internal")
        self.assertEqual(len(tunnels), 1)
        self.assertEqual(tunnels[0].config["addr"], "http://localhost:80")
        self.assertTrue(tunnels[0].config["addr"].startswith("http://"))
        self.assertEqual(tunnels[0].proto, "https")
        self.assertEqual(tunnels[0].public_url, "https://pyngrok.internal")

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_ngrok_tls_internal_endpoint_pooling(self):
        # WHEN
        tls_internal_endpoint = ngrok.connect(proto="tls", addr="443", domain="pyngrok.internal",
                                              pooling_enabled=True,
                                              pyngrok_config=self.pyngrok_config_v3)
        tunnels = ngrok.get_tunnels(pyngrok_config=self.pyngrok_config_v3)

        # THEN
        self.assertEqual(tls_internal_endpoint.config["addr"], "tls://localhost:443")
        self.assertTrue(tls_internal_endpoint.config["addr"].startswith("tls://"))
        self.assertEqual(tls_internal_endpoint.proto, "tls")
        self.assertEqual(tls_internal_endpoint.public_url, "tls://pyngrok.internal")
        self.assertEqual(len(tunnels), 1)
        self.assertEqual(tunnels[0].config["addr"], "tls://localhost:443")
        self.assertTrue(tunnels[0].config["addr"].startswith("tls://"))
        self.assertEqual(tunnels[0].proto, "tls")
        self.assertEqual(tunnels[0].public_url, "tls://pyngrok.internal")

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_tunnel_definitions_pyngrok_default(self):
        subdomain = generate_name_for_subdomain("pyngrok-temp")

        # GIVEN
        config = {
            "tunnels": {
                "pyngrok-default": {
                    "proto": "http",
                    "addr": "8080",
                    "subdomain": subdomain
                }
            }
        }
        config_path = os.path.join(self.config_dir, "config_v3_2.yml")
        installer.install_default_config(config_path, config, ngrok_version="v3")
        subdomain = generate_name_for_subdomain("pyngrok-temp")
        pyngrok_config = self.copy_with_updates(self.pyngrok_config_v3,
                                                config_path=config_path)

        # WHEN
        ngrok_tunnel = ngrok.connect(pyngrok_config=pyngrok_config)

        # THEN
        self.assertEqual(ngrok_tunnel.name, "pyngrok-default-api")
        self.assertEqual(ngrok_tunnel.config["addr"],
                         f"http://localhost:{config['tunnels']['pyngrok-default']['addr']}")
        self.assertEqual("http", config["tunnels"]["pyngrok-default"]["proto"])
        self.assertEqual(ngrok_tunnel.public_url,
                         f"https://{config['tunnels']['pyngrok-default']['subdomain']}.ngrok.io")

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_tunnel_definitions_pyngrok_default_with_overrides(self):
        subdomain = generate_name_for_subdomain("pyngrok-temp")

        # GIVEN
        config = {
            "tunnels": {
                "pyngrok-default": {
                    "proto": "http",
                    "addr": "8080",
                    "subdomain": subdomain
                }
            }
        }
        config_path = os.path.join(self.config_dir, "config_v3_2.yml")
        installer.install_default_config(config_path, config, ngrok_version="v3")
        subdomain = generate_name_for_subdomain("pyngrok-temp")
        pyngrok_config = self.copy_with_updates(self.pyngrok_config_v3,
                                                config_path=config_path)

        # WHEN
        ngrok_tunnel = ngrok.connect("5000", subdomain=subdomain, pyngrok_config=pyngrok_config)

        # THEN
        self.assertEqual(ngrok_tunnel.name, "pyngrok-default-api")
        self.assertEqual(ngrok_tunnel.config["addr"], "http://localhost:5000")
        self.assertEqual("http", config["tunnels"]["pyngrok-default"]["proto"])
        self.assertIn(subdomain, ngrok_tunnel.public_url)

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_upgrade_ngrok_config_file_v2_to_v3(self):
        # GIVEN
        config_path = os.path.join(self.config_dir, "legacy_config.yml")
        installer.install_default_config(config_path, ngrok_version="v2")
        pyngrok_config_v3 = self.copy_with_updates(self.pyngrok_config_v3,
                                                   config_path=config_path,
                                                   ngrok_version="v3")
        self.given_ngrok_installed(self.pyngrok_config_v3)

        # WHEN
        with self.assertRaises(PyngrokError):
            ngrok.connect(pyngrok_config=pyngrok_config_v3)

        # GIVEN
        process.capture_run_process(pyngrok_config_v3.ngrok_path,
                                    ["config", "upgrade", "--config", pyngrok_config_v3.config_path])

        # WHEN
        ngrok.connect(pyngrok_config=pyngrok_config_v3)

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_full_config_v2_http_tunnel_definitions(self):
        # GIVEN
        config = {
            "version": "2",
            "tunnels": {
                "my-tunnel": {
                    "addr": "5000",
                    "metadata": "metadata",
                    "inspect": False,
                    "proto": "http",
                    "basic_auth": ["auth-token"],
                    "circuit_breaker": 20,
                    "compression": False,
                    "host_header": "host-header",
                    "domain": "pyngrok.com",
                    "ip_restriction": {"allow_cidrs": ["allowed"], "deny_cidrs": ["denied"]},
                    "mutual_tls_cas": __file__,
                    "oauth": {
                        "provider": "google",
                        "allow_domains": ["pyngrok.com"],
                        "allow_emails": ["email@pyngrok.com"]
                    },
                    "proxy_proto": "1",
                    "request_header": {"add": ["req-addition"], "remove": ["req-subtraction"]},
                    "response_header": {"add": ["res-addition"], "remove": ["res-subtraction"]},
                    "schemes": ["https"],
                    # Can't be used in conjunction with "domain", but validated in other tests
                    # "subdomain": "pyngrok",
                    "traffic_policy": {
                        "on_http_request": {"name": "inbound-policy", "expressions": "inbound-policy-expression",
                                            "actions": {"type": "inbound-policy-actions-type",
                                                        "config": "inbound-policy-actions-config"}},
                        "on_http_response": {"name": "outbound-policy", "expressions": "outbound-policy-expression",
                                             "actions": {"type": "outbound-policy-actions-type",
                                                         "config": "outbound-policy-actions-config"}}
                    },
                    "user_agent_filter": {"allow": ["allow-user-agent"], "deny": ["deny-user-agent"]},
                    "verify_webhook": {"provider": "provider", "secret": "secret"},
                    "websocket_tcp_converter": False
                }
            }
        }
        config_path = os.path.join(self.config_dir, "config_v3_2.yml")
        installer.install_default_config(config_path, config, ngrok_version="v3")
        pyngrok_config = self.copy_with_updates(self.pyngrok_config_v3,
                                                config_path=config_path,
                                                api_key="api-key")

        # WHEN
        # This test ensures the validity of the config (pyngrok will fail with PyngrokNgrokError if the above is
        # invalid), not that the config can start a valid tunnel (thus PyngrokNgrokHTTPError is thrown when pyngrok
        # tries to open the tunnel after ngrok successfully starts, but we can ignore that for this test
        with self.assertRaises(PyngrokNgrokHTTPError):
            ngrok.connect(name="my-tunnel", pyngrok_config=pyngrok_config)

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_full_config_v2_tcp_tunnel_definitions(self):
        # GIVEN
        config = {
            "version": "2",
            "tunnels": {
                "my-tunnel": {
                    "addr": "5000",
                    "metadata": "metadata",
                    "inspect": False,
                    "proto": "tcp",
                    "ip_restriction": {"allow_cidrs": ["allowed"], "deny_cidrs": ["denied"]},
                    "proxy_proto": "1",
                    "remote_addr": "2.tcp.ngrok.io:21746",
                    "traffic_policy": {
                        "inbound": {"name": "inbound-policy", "expressions": "inbound-policy-expression",
                                    "actions": {"type": "inbound-policy-actions-type",
                                                "config": "inbound-policy-actions-config"}},
                        "outbound": {"name": "outbound-policy", "expressions": "outbound-policy-expression",
                                     "actions": {"type": "outbound-policy-actions-type",
                                                 "config": "outbound-policy-actions-config"}}
                    }
                }
            }
        }
        config_path = os.path.join(self.config_dir, "config_v3_2.yml")
        installer.install_default_config(config_path, config, ngrok_version="v3")
        pyngrok_config = self.copy_with_updates(self.pyngrok_config_v3,
                                                config_path=config_path,
                                                api_key="api-key")

        # WHEN
        # This test ensures the validity of the config (pyngrok will fail with PyngrokNgrokError if the above is
        # invalid), not that the config can start a valid tunnel (thus PyngrokNgrokHTTPError is thrown when pyngrok
        # tries to open the tunnel after ngrok successfully starts, but we can ignore that for this test
        with self.assertRaises(PyngrokNgrokHTTPError):
            ngrok.connect(name="my-tunnel", pyngrok_config=pyngrok_config)

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_full_config_v2_tls_tunnel_definitions(self):
        # GIVEN
        config = {
            "version": "2",
            "tunnels": {
                "my-tunnel": {
                    "addr": "5000",
                    "metadata": "metadata",
                    "inspect": False,
                    "proto": "tls",
                    # "mutual_tls_cas": __file__,
                    # "crt": __file__,
                    "domain": "pyngrok.com",
                    "ip_restriction": {"allow_cidrs": ["allowed"], "deny_cidrs": ["denied"]},
                    # "key": __file__,
                    "proxy_proto": "1",
                    # Can't be used in conjunction with "domain", but validated in other tests
                    # "subdomain": "pyngrok",
                    "terminate_at": "edge",
                    "traffic_policy": {
                        "inbound": {"name": "inbound-policy", "expressions": "inbound-policy-expression",
                                    "actions": {"type": "inbound-policy-actions-type",
                                                "config": "inbound-policy-actions-config"}},
                        "outbound": {"name": "outbound-policy", "expressions": "outbound-policy-expression",
                                     "actions": {"type": "outbound-policy-actions-type",
                                                 "config": "outbound-policy-actions-config"}}
                    }
                }
            }
        }
        config_path = os.path.join(self.config_dir, "config_v3_2.yml")
        installer.install_default_config(config_path, config, ngrok_version="v3")
        pyngrok_config = self.copy_with_updates(self.pyngrok_config_v3,
                                                config_path=config_path,
                                                api_key="api-key")

        # WHEN
        # This test ensures the validity of the config (pyngrok will fail with PyngrokNgrokError if the above is
        # invalid), not that the config can start a valid tunnel (thus PyngrokNgrokHTTPError is thrown when pyngrok
        # tries to open the tunnel after ngrok successfully starts, but we can ignore that for this test
        with self.assertRaises(PyngrokNgrokHTTPError):
            ngrok.connect(name="my-tunnel", pyngrok_config=pyngrok_config)

    ################################################################################
    # Tests below this point don't need to start a long-lived ngrok process, they
    # are asserting on pyngrok-specific code or edge cases.
    ################################################################################

    def test_web_addr_false_not_allowed(self):
        # GIVEN
        with open(self.pyngrok_config_v3.config_path, "w") as config_file:
            yaml.dump({"web_addr": False}, config_file)

        # WHEN
        with self.assertRaises(PyngrokError):
            ngrok.connect(pyngrok_config=self.pyngrok_config_v3)

    def test_log_format_json_not_allowed(self):
        # GIVEN
        with open(self.pyngrok_config_v3.config_path, "w") as config_file:
            yaml.dump({"log_format": "json"}, config_file)

        # WHEN
        with self.assertRaises(PyngrokError):
            ngrok.connect(pyngrok_config=self.pyngrok_config_v3)

    def test_log_level_warn_not_allowed(self):
        # GIVEN
        with open(self.pyngrok_config_v3.config_path, "w") as config_file:
            yaml.dump({"log_level": "warn"}, config_file)

        # WHEN
        with self.assertRaises(PyngrokError):
            ngrok.connect(pyngrok_config=self.pyngrok_config_v3)

    def test_api_request_security_error(self):
        # WHEN
        with self.assertRaises(PyngrokSecurityError):
            ngrok.api_request(f"file:{__file__}")

    @mock.patch("pyngrok.process.capture_run_process")
    def test_update(self, mock_capture_run_process):
        ngrok.update(pyngrok_config=self.pyngrok_config_v3)

        self.assertEqual(mock_capture_run_process.call_count, 1)
        self.assertEqual("update", mock_capture_run_process.call_args[0][1][0])

    def test_version(self):
        # WHEN
        ngrok_version, pyngrok_version = ngrok.get_version(pyngrok_config=self.pyngrok_config_v3)

        # THEN
        self.assertIsNotNone(ngrok_version)
        self.assertEqual(__version__, pyngrok_version)

    def test_set_auth_token_v2(self):
        ngrok.set_auth_token("some-auth-token", pyngrok_config=self.pyngrok_config_v2)
        with open(self.pyngrok_config_v2.config_path, "r") as f:
            contents = f.read()

        # THEN
        self.assertIn("some-auth-token", contents)

    def test_set_auth_token_v3(self):
        # WHEN
        ngrok.set_auth_token("some-auth-token", pyngrok_config=self.pyngrok_config_v3)
        with open(self.pyngrok_config_v3.config_path, "r") as f:
            contents = f.read()

        # THEN
        self.assertIn("some-auth-token", contents)

    @mock.patch("subprocess.check_output")
    def test_set_auth_token_fails(self, mock_check_output):
        # GIVEN
        error_msg = "An error occurred"
        mock_check_output.return_value = error_msg

        # WHEN
        with self.assertRaises(PyngrokNgrokError) as cm:
            ngrok.set_auth_token("some-auth-token", pyngrok_config=self.pyngrok_config_v3)

        # THEN
        self.assertIn(f": {error_msg}", str(cm.exception))

    def test_set_api_key_v2_fails(self):
        with self.assertRaises(PyngrokError):
            ngrok.set_api_key("some-api-key", pyngrok_config=self.pyngrok_config_v2)

    def test_set_api_key_v3(self):
        # WHEN
        ngrok.set_api_key("some-api-key", pyngrok_config=self.pyngrok_config_v3)
        with open(self.pyngrok_config_v3.config_path, "r") as f:
            contents = f.read()

        # THEN
        self.assertIn("some-api-key", contents)

    @mock.patch("subprocess.check_output")
    def test_set_api_key_fails(self, mock_check_output):
        # GIVEN
        error_msg = "An error occurred"
        mock_check_output.return_value = error_msg

        # WHEN
        with self.assertRaises(PyngrokNgrokError) as cm:
            ngrok.set_api_key("some-api-key", pyngrok_config=self.pyngrok_config_v3)

        # THEN
        self.assertIn(f": {error_msg}", str(cm.exception))

    @mock.patch('pyngrok.ngrok.api_request')
    @mock.patch('pyngrok.ngrok.get_ngrok_process')
    def test_config_v2_passed_to_api_request(self, mock_get_ngrok_process, mock_api_request):
        # GIVEN
        config = {
            "version": "2",
            "tunnels": {
                "my-tunnel": {
                    "proto": "http",
                    "addr": "8000"
                }
            }
        }
        config_path = os.path.join(self.config_dir, "config_v3_2.yml")

        installer.install_default_config(config_path, config, ngrok_version="v3", config_version="3")
        pyngrok_config = self.copy_with_updates(self.pyngrok_config_v3,
                                                config_path=config_path,
                                                api_key="api-key", )

        expected_options = config["tunnels"]["my-tunnel"].copy()
        expected_options["name"] = "my-tunnel-api"

        # WHEN
        ngrok.connect(name="my-tunnel", pyngrok_config=pyngrok_config)

        # THEN
        mock_api_request.assert_called_with(f"{mock_get_ngrok_process().api_url}/api/tunnels", method="POST",
                                            data=expected_options, timeout=pyngrok_config.request_timeout)



__copyright__ = "Copyright (c) 2018-2025 Alex Laird"
__license__ = "MIT"

import copy
import logging
import os
import platform
import socket
import sys
import tempfile
import threading
import time
import zipfile
from http import HTTPStatus
from typing import Any, Dict, Optional
from urllib.error import URLError
from urllib.request import urlopen

import yaml

from pyngrok.exception import PyngrokError, PyngrokNgrokInstallError, PyngrokSecurityError

logger = logging.getLogger(__name__)

CDN_URL_PREFIX = "https://bin.equinox.io/c/4VmDzA7iaHb/"
CDN_V3_URL_PREFIX = "https://bin.equinox.io/c/bNyj1mQVY4c/"
PLATFORMS = {
    "darwin_x86_64": CDN_URL_PREFIX + "ngrok-stable-darwin-amd64.zip",
    "darwin_x86_64_arm": CDN_URL_PREFIX + "ngrok-stable-darwin-arm64.zip",
    "windows_i386": CDN_URL_PREFIX + "ngrok-stable-windows-386.zip",
    "windows_x86_64": CDN_URL_PREFIX + "ngrok-stable-windows-amd64.zip",
    "linux_i386": CDN_URL_PREFIX + "ngrok-stable-linux-386.zip",
    "linux_i386_arm": CDN_URL_PREFIX + "ngrok-stable-linux-arm.zip",
    "linux_x86_64": CDN_URL_PREFIX + "ngrok-stable-linux-amd64.zip",
    "linux_x86_64_arm": CDN_URL_PREFIX + "ngrok-stable-linux-arm64.zip",
    "freebsd_i386": CDN_URL_PREFIX + "ngrok-stable-freebsd-386.zip",
    "freebsd_x86_64": CDN_URL_PREFIX + "ngrok-stable-freebsd-amd64.zip"
}
PLATFORMS_V3 = {
    "darwin_x86_64": CDN_V3_URL_PREFIX + "ngrok-v3-stable-darwin-amd64.zip",
    "darwin_x86_64_arm": CDN_V3_URL_PREFIX + "ngrok-v3-stable-darwin-arm64.zip",
    "windows_i386": CDN_V3_URL_PREFIX + "ngrok-v3-stable-windows-386.zip",
    "windows_x86_64": CDN_V3_URL_PREFIX + "ngrok-v3-stable-windows-amd64.zip",
    "windows_x86_64_arm": CDN_V3_URL_PREFIX + "ngrok-v3-stable-windows-arm64.zip",
    "linux_i386": CDN_V3_URL_PREFIX + "ngrok-v3-stable-linux-386.zip",
    "linux_i386_arm": CDN_V3_URL_PREFIX + "ngrok-v3-stable-linux-arm.zip",
    "linux_x86_64": CDN_V3_URL_PREFIX + "ngrok-v3-stable-linux-amd64.zip",
    "linux_x86_64_arm": CDN_V3_URL_PREFIX + "ngrok-v3-stable-linux-arm64.zip",
    "linux_s390x": CDN_V3_URL_PREFIX + "ngrok-v3-stable-linux-s390x.zip",
    "linux_ppc64": CDN_V3_URL_PREFIX + "ngrok-v3-stable-linux-ppc64.zip",
    "linux_ppc64le": CDN_V3_URL_PREFIX + "ngrok-v3-stable-linux-ppc64le.zip",
    "freebsd_i386": CDN_V3_URL_PREFIX + "ngrok-v3-stable-freebsd-386.zip",
    "freebsd_x86_64": CDN_V3_URL_PREFIX + "ngrok-v3-stable-freebsd-amd64.zip",
    "freebsd_i386_arm": CDN_V3_URL_PREFIX + "ngrok-v3-stable-freebsd-arm.zip"
}
UNIX_BINARIES = ["darwin", "linux", "freebsd"]
SUPPORTED_NGROK_VERSIONS = ["v2", "v3"]
DEFAULT_DOWNLOAD_TIMEOUT = 6
DEFAULT_RETRY_COUNT = 0

config_file_lock = threading.RLock()

_config_cache: Dict[str, Dict[str, Any]] = {}
_print_progress_enabled = True


def get_default_ngrok_dir() -> str:
    """
    Get the default ``ngrok`` directory for the current system.

    :return: The default ``ngrok`` directory.
    """
    system = get_system()
    user_home = os.path.expanduser("~")
    if system == "darwin":
        return os.path.join(user_home, "Library", "Application Support", "ngrok")
    elif system == "windows":
        return os.path.join(user_home, "AppData", "Local", "ngrok")
    else:
        return os.path.join(os.environ.get("XDG_CONFIG_HOME", os.path.join(user_home, ".config")), "ngrok")


def get_system() -> str:
    """
    Parse the name of the OS from the system and return a friendly name.

    :return: The friendly name of the OS.
    :raises: :class:`~pyngrok.exception.PyngrokNgrokInstallError`: When the platform is not supported.
    """
    system = platform.system().replace(" ", "").lower()

    if system.startswith("darwin"):
        return "darwin"
    elif system.startswith("windows") \
            or system.startswith("cygwin") \
            or system.startswith("ming"):
        return "windows"
    elif system.startswith("linux"):
        return "linux"
    elif system.startswith("freebsd"):
        return "freebsd"
    else:
        raise PyngrokNgrokInstallError(f"\"{system}\" is not a supported system")


def get_arch() -> str:
    """
    Get the architecture of the current system. This will be ``i386`` for 32-bit systems, ``x86_64`` for 64-bit
    systems, and have ``_arm`` appended for ARM systems. Special architectures like `s390x` and `ppc64le` will be
    returned as-is.

    :return: The name of the architecture.
    """
    machine = platform.machine().lower()

    if machine in ["s390x", "ppc64", "ppc64le"]:
        return machine

    arch = "x86_64" if machine.endswith("64") else "i386"
    if machine.startswith("arm") or \
            machine.startswith("aarch64"):
        arch += "_arm"
    return arch


def get_ngrok_bin() -> str:
    """
    Get the ``ngrok`` executable for the current system.

    :return: The name of the ``ngrok`` executable.
    :raises: :class:`~pyngrok.exception.PyngrokNgrokInstallError`: When the platform is not supported.
    """
    system = get_system()
    if system in UNIX_BINARIES:
        return "ngrok"
    else:
        return "ngrok.exe"


def get_ngrok_cdn_url(ngrok_version: Optional[str]) -> str:
    """
    Determine the ``ngrok`` CDN URL for the current OS and architecture.

    :param ngrok_version: The major version of ``ngrok`` to be installed.
    :return: The ``ngrok`` CDN URL.
    """
    plat = "{system}_{arch}".format(system=get_system(), arch=get_arch())

    logger.debug(f"Platform to download: {plat}")

    try:
        if ngrok_version == "v2":
            return PLATFORMS[plat]
        elif ngrok_version == "v3":
            return PLATFORMS_V3[plat]
        else:
            raise PyngrokError(f"\"ngrok_version\" must be a supported version: {SUPPORTED_NGROK_VERSIONS}")
    except KeyError:
        raise PyngrokNgrokInstallError(f"\"{plat}\" is not a supported platform")


def install_ngrok(ngrok_path: str,
                  ngrok_version: Optional[str] = "v3",
                  **kwargs: Any) -> None:
    """
    Download and install the latest ``ngrok`` for the current system, overwriting any existing contents
    at the given path.

    :param ngrok_path: The path to where the ``ngrok`` binary will be downloaded.
    :param ngrok_version: The major version of ``ngrok`` to be installed.
    :param kwargs: Remaining ``kwargs`` will be passed to :func:`_download_file`.
    :raises: :class:`~pyngrok.exception.PyngrokError`: When the ``ngrok_version`` is not supported.
    :raises: :class:`~pyngrok.exception.PyngrokNgrokInstallError`: When an error occurs installing ``ngrok``.
    """
    logger.debug(
        "Installing ngrok {ngrok_version} to "
        "{ngrok_path}{optional_overwrite} ...".format(ngrok_version=ngrok_version,
                                                      ngrok_path=ngrok_path,
                                                      optional_overwrite=", overwriting" if os.path.exists(
                                                          ngrok_path) else ""))

    ngrok_dir = os.path.dirname(ngrok_path)

    if not os.path.exists(ngrok_dir):
        os.makedirs(ngrok_dir)

    url = get_ngrok_cdn_url(ngrok_version)

    try:
        download_path = _download_file(url, **kwargs)

        _install_ngrok_zip(ngrok_path, download_path)
    except Exception as e:
        raise PyngrokNgrokInstallError(f"An error occurred while downloading ngrok from {url}: {e}")


def _install_ngrok_zip(ngrok_path: str,
                       zip_path: str) -> None:
    """
    Extract the ``ngrok`` zip file to the given path.

    :param ngrok_path: The path where ``ngrok`` will be installed.
    :param zip_path: The path to the ``ngrok`` zip file to be extracted.
    """
    _print_progress("Installing ngrok ... ")

    with zipfile.ZipFile(zip_path, "r") as zip_ref:
        logger.debug(f"Extracting ngrok binary from {zip_path} to {ngrok_path} ...")
        zip_ref.extractall(os.path.dirname(ngrok_path))

    os.chmod(ngrok_path, int("700", 8))

    _clear_progress()


def get_ngrok_config(config_path: str,
                     use_cache: bool = True,
                     ngrok_version: Optional[str] = "v3",
                     config_version: Optional[str] = "2") -> Dict[str, Any]:
    """
    Get the ``ngrok`` config from the given path.

    :param config_path: The ``ngrok`` config path to read.
    :param use_cache: Use the cached version of the config (if populated).
    :param ngrok_version: The major version of ``ngrok`` installed.
    :param config_version: The ``ngrok`` config version.
    :return: The ``ngrok`` config.
    """
    with config_file_lock:
        if config_path not in _config_cache or not use_cache:
            with open(config_path, "r") as config_file:
                config = yaml.safe_load(config_file)
                if config is None:
                    config = get_default_config(ngrok_version, config_version)

            _config_cache[config_path] = config

    return _config_cache[config_path]


def get_default_config(ngrok_version: Optional[str],
                       config_version: Optional[str]) -> Dict[str, Any]:
    """
    Get the default config params for the given major version of ``ngrok`` and config version.

    :param ngrok_version: The major version of ``ngrok`` installed.
    :param config_version: The ``ngrok`` config version.
    :return: The default config.
    :raises: :class:`~pyngrok.exception.PyngrokError`: When the ``ngrok_version`` is not supported.
    """
    if ngrok_version == "v2":
        return {}
    elif ngrok_version == "v3":
        config = {"version": config_version}
        if str(config_version) == "2":
            config["region"] = "us"
        return config
    else:
        raise PyngrokError(f"\"ngrok_version\" must be a supported version: {SUPPORTED_NGROK_VERSIONS}")


def install_default_config(config_path: str,
                           data: Optional[Dict[str, Any]] = None,
                           ngrok_version: Optional[str] = "v3",
                           config_version: Optional[str] = "2") -> None:
    """
    Install the given data to the ``ngrok`` config. If a config is not already present for the given path, create one.
    Before saving new data to the default config, validate that they are compatible with ``pyngrok``.

    :param config_path: The path to where the ``ngrok`` config should be installed.
    :param data: A dictionary of things to add to the default config.
    :param ngrok_version: The major version of ``ngrok`` installed.
    :param config_version: The ``ngrok`` config version.
    """
    with config_file_lock:
        if data is None:
            data = {}
        else:
            data = copy.deepcopy(data)

        data.update(get_default_config(ngrok_version, config_version))

        config_dir = os.path.dirname(config_path)
        if not os.path.exists(config_dir):
            os.makedirs(config_dir)
        if not os.path.exists(config_path):
            open(config_path, "w").close()

        config = get_ngrok_config(config_path,
                                  use_cache=False,
                                  ngrok_version=ngrok_version,
                                  config_version=config_version)

        config.update(data)

        validate_config(config)

        with open(config_path, "w") as config_file:
            logger.debug(f"Installing default ngrok config to {config_path} ...")

            yaml.dump(config, config_file)


def validate_config(data: Dict[str, Any]) -> None:
    """
    Validate that the given dict of config items are valid for ``ngrok`` and ``pyngrok``.

    :param data: A dictionary of things to be validated as config items.
    :raises: :class:`~pyngrok.exception.PyngrokError`: When a key or value fails validation.
    """
    if data.get("web_addr", None) is False:
        raise PyngrokError("\"web_addr\" cannot be False, as the ngrok API is a dependency for pyngrok")
    elif data.get("log_format") == "json":
        raise PyngrokError("\"log_format\" must be \"term\" to be compatible with pyngrok")
    elif data.get("log_level", "info") not in ["info", "debug"]:
        raise PyngrokError("\"log_level\" must be \"info\" to be compatible with pyngrok")


def _download_file(url: str,
                   retries: int = 0,
                   **kwargs: Any) -> str:
    """
    Download a file to a temporary path and emit a status to stdout (if possible) as the download progresses.

    :param url: The URL to download.
    :param retries: The retry attempt index, if download fails.
    :param kwargs: Remaining ``kwargs`` will be passed to :py:func:`urllib.request.urlopen`.
    :return: The path to the downloaded temporary file.
    :raises: :class:`~pyngrok.exception.PyngrokSecurityError`: When the ``url`` is not supported.
    :raises: :class:`~pyngrok.exception.PyngrokNgrokInstallError`: When an error occurs downloading ``ngrok``.
    """
    kwargs["timeout"] = kwargs.get("timeout", DEFAULT_DOWNLOAD_TIMEOUT)

    if not url.lower().startswith("http"):
        raise PyngrokSecurityError(f"URL must start with \"http\": {url}")

    try:
        _print_progress("Downloading ngrok ...")

        logger.debug(f"Download ngrok from {url} ...")

        local_filename = url.split("/")[-1]
        response = urlopen(url, **kwargs)

        status_code = response.getcode()

        if status_code != HTTPStatus.OK:
            logger.debug(f"Response status code: {status_code}")

            raise PyngrokNgrokInstallError(f"Download failed, status code: {status_code}")

        length = response.getheader("Content-Length")
        if length:
            length = int(length)
            chunk_size = max(4096, length // 100)
        else:
            chunk_size = 64 * 1024

        download_path = os.path.join(tempfile.gettempdir(), local_filename)
        with open(download_path, "wb") as f:
            size = 0
            while True:
                buffer = response.read(chunk_size)

                if not buffer:
                    break

                f.write(buffer)
                size += len(buffer)

                if length:
                    percent_done = int((float(size) / float(length)) * 100)
                    _print_progress(f"Downloading ngrok: {percent_done}%")

        _clear_progress()

        return download_path
    except (socket.timeout, URLError) as e:
        if retries < DEFAULT_RETRY_COUNT:
            logger.warning("ngrok download failed, retrying in 0.5 seconds ...")
            time.sleep(0.5)

            return _download_file(url, retries + 1, **kwargs)
        else:
            raise e


def _print_progress(line: str) -> None:
    if _print_progress_enabled:
        sys.stdout.write(f"{line}\r")
        sys.stdout.flush()


def _clear_progress(spaces: int = 100) -> None:
    if _print_progress_enabled:
        sys.stdout.write((" " * spaces) + "\r")
        sys.stdout.flush()



__copyright__ = "Copyright (c) 2018-2024 Alex Laird"
__license__ = "MIT"

from typing import Any, List, Optional, Union, Dict, MutableMapping

from pyngrok.log import NgrokLog


class PyngrokError(Exception):
    """
    Raised when a general ``pyngrok`` error has occurred.
    """
    pass


class PyngrokSecurityError(PyngrokError):
    """
    Raised when a ``pyngrok`` security error has occurred.
    """
    pass


class PyngrokNgrokInstallError(PyngrokError):
    """
    Raised when an error has occurred while downloading and installing the ``ngrok`` binary.
    """
    pass


class PyngrokNgrokError(PyngrokError):
    """
    Raised when an error occurs interacting directly with the ``ngrok`` binary.
    """

    def __init__(self,
                 error: Union[str, BaseException],
                 ngrok_logs: Optional[List[NgrokLog]] = None,
                 ngrok_error: Optional[str] = None) -> None:
        super(PyngrokNgrokError, self).__init__(error)

        #: The ``ngrok`` logs, which may be useful for debugging the error.
        self.ngrok_logs: List[NgrokLog] = ngrok_logs if ngrok_logs else []
        #: The error that caused the ``ngrok`` process to fail.
        self.ngrok_error: Optional[str] = ngrok_error


class PyngrokNgrokHTTPError(PyngrokNgrokError):
    """
    Raised when an error occurs making a request to the ``ngrok`` web interface. The ``body``
    contains the error response received from ``ngrok``.
    """

    def __init__(self,
                 error: Union[str, BaseException],
                 url: str,
                 status_code: int,
                 message: Optional[str],
                 headers: Union[Dict[str, str], MutableMapping[str, str], Any],
                 body: str) -> None:
        super(PyngrokNgrokHTTPError, self).__init__(error)

        #: The request URL that failed.
        self.url: str = url
        #: The response status code from ``ngrok``.
        self.status_code: int = status_code
        #: The response message from ``ngrok``.
        self.message: Optional[str] = message
        #: The request headers sent to ``ngrok``.
        self.headers: Any = headers
        #: The response body from ``ngrok``.
        self.body: str = body


class PyngrokNgrokURLError(PyngrokNgrokError):
    """
    Raised when an error occurs when trying to initiate an API request.
    """

    def __init__(self,
                 error: Union[str, BaseException],
                 reason: Any) -> None:
        super(PyngrokNgrokURLError, self).__init__(error)

        #: The reason for the URL error.
        self.reason: Any = reason



__copyright__ = "Copyright (c) 2018-2024 Alex Laird"
__license__ = "MIT"

import os
import platform
import time
import unittest
from unittest import mock
from urllib.parse import urlparse
from urllib.request import urlopen

from pyngrok import installer, ngrok, process
from pyngrok.exception import PyngrokNgrokError
from pyngrok.process import NgrokLog
from tests.testcase import NgrokTestCase


class TestProcess(NgrokTestCase):
    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_terminate_process(self):
        # GIVEN
        self.given_ngrok_installed(self.pyngrok_config_v3)
        ngrok_process = process._start_process(self.pyngrok_config_v3)
        monitor_thread = ngrok_process._monitor_thread
        self.assertIsNone(ngrok_process.proc.poll())

        # WHEN
        process._terminate_process(ngrok_process.proc)
        time.sleep(1)

        # THEN
        self.assertIsNotNone(ngrok_process.proc.poll())
        self.assertFalse(monitor_thread.is_alive())

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_start_process_port_in_use_v2(self):
        # GIVEN
        self.given_ngrok_installed(self.pyngrok_config_v2)
        self.assertEqual(len(process._current_processes.keys()), 0)
        ngrok_process = process._start_process(self.pyngrok_config_v2)
        port = urlparse(ngrok_process.api_url).port
        self.assertEqual(len(process._current_processes.keys()), 1)
        self.assertTrue(ngrok_process._monitor_thread.is_alive())

        ngrok_path_v2_2 = os.path.join(self.config_dir, "v2_2", installer.get_ngrok_bin())
        pyngrok_config_v2_2 = self.copy_with_updates(self.pyngrok_config_v2,
                                                     ngrok_path=ngrok_path_v2_2)
        self.given_ngrok_installed(pyngrok_config_v2_2)
        config_path_v2_2 = os.path.join(self.config_dir, "config_v2_2.yml")
        installer.install_default_config(config_path_v2_2, {"web_addr": ngrok_process.api_url.lstrip("http://")},
                                         ngrok_version="v2")
        pyngrok_config_v2_2 = self.copy_with_updates(self.pyngrok_config_v2,
                                                     config_path=config_path_v2_2,
                                                     ngrok_path=ngrok_path_v2_2)

        time.sleep(3)

        # WHEN
        with self.assertRaises(PyngrokNgrokError) as cm:
            process._start_process(pyngrok_config_v2_2)

        # THEN
        self.assertIsNotNone(cm.exception)
        if platform.system() == "Windows":
            self.assertIn(f"{port}: bind: Only one usage of each socket address", cm.exception.ngrok_error)
        else:
            self.assertIn(f"{port}: bind: address already in use", str(cm.exception.ngrok_error))
        self.assertEqual(len(process._current_processes.keys()), 1)
        self.assert_no_zombies()

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_start_process_port_in_use_v3(self):
        # GIVEN
        self.given_ngrok_installed(self.pyngrok_config_v3)
        self.assertEqual(len(process._current_processes.keys()), 0)
        ngrok_process = process._start_process(self.pyngrok_config_v3)
        port = urlparse(ngrok_process.api_url).port
        self.assertEqual(len(process._current_processes.keys()), 1)
        self.assertTrue(ngrok_process._monitor_thread.is_alive())

        ngrok_path_v3_2 = os.path.join(self.config_dir, "v3_2", installer.get_ngrok_bin())
        pyngrok_config_v3_2 = self.copy_with_updates(self.pyngrok_config_v3,
                                                     ngrok_path=ngrok_path_v3_2)
        self.given_ngrok_installed(pyngrok_config_v3_2)
        config_path_v3_2 = os.path.join(self.config_dir, "config_v3_2.yml")
        installer.install_default_config(config_path_v3_2, {"web_addr": ngrok_process.api_url.lstrip("http://")},
                                         ngrok_version="v3")
        pyngrok_config_v3_2 = self.copy_with_updates(self.pyngrok_config_v3,
                                                     config_path=config_path_v3_2,
                                                     ngrok_path=ngrok_path_v3_2)

        time.sleep(3)

        # WHEN
        with self.assertRaises(PyngrokNgrokError) as cm:
            process._start_process(pyngrok_config_v3_2)

        # THEN
        self.assertIsNotNone(cm.exception.ngrok_error)
        if platform.system() == "Windows":
            self.assertIn(f"{port}: bind: Only one usage of each socket address", cm.exception.ngrok_error)
        else:
            self.assertIn(f"{port}: bind: address already in use", str(cm.exception.ngrok_error))
        self.assertEqual(len(process._current_processes.keys()), 1)
        self.assert_no_zombies()

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_process_external_kill(self):
        # GIVEN
        self.given_ngrok_installed(self.pyngrok_config_v3)
        ngrok_process1 = process._start_process(self.pyngrok_config_v3)
        monitor_thread1 = ngrok_process1._monitor_thread
        self.assertEqual(len(process._current_processes.keys()), 1)
        self.assertTrue(ngrok_process1._monitor_thread.is_alive())

        # WHEN
        # Kill the process by external means, pyngrok still thinks process is active
        ngrok_process1.proc.kill()
        ngrok_process1.proc.wait()
        time.sleep(1)
        self.assertEqual(len(process._current_processes.keys()), 1)
        self.assertFalse(monitor_thread1.is_alive())

        # THEN
        # Try to kill the process via pyngrok, no error, just update state
        process.kill_process(self.pyngrok_config_v3.ngrok_path)
        self.assertEqual(len(process._current_processes.keys()), 0)
        self.assertFalse(monitor_thread1.is_alive())
        self.assert_no_zombies()

        # THEN
        # Try to get process via pyngrok, it has been killed, restart and correct state
        with mock.patch("atexit.register") as mock_atexit:
            ngrok_process2 = process.get_process(self.pyngrok_config_v3)
            self.assertNotEqual(ngrok_process1.proc.pid, ngrok_process2.proc.pid)
            self.assertEqual(len(process._current_processes.keys()), 1)
            self.assertFalse(monitor_thread1.is_alive())
            self.assertTrue(ngrok_process2._monitor_thread.is_alive())

            self.assertTrue(mock_atexit.called)
            self.assert_no_zombies()

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_multiple_processes_different_binaries(self):
        # GIVEN
        self.assertEqual(len(process._current_processes.keys()), 0)

        self.given_ngrok_installed(self.pyngrok_config_v2)
        config_path_v2_1 = os.path.join(self.config_dir, "config_v2_1.yml")
        installer.install_default_config(config_path_v2_1, {"web_addr": "localhost:4040"}, ngrok_version="v2")
        pyngrok_config_v2_1 = self.copy_with_updates(self.pyngrok_config_v2,
                                                     config_path=config_path_v2_1,
                                                     ngrok_path=self.pyngrok_config_v2.ngrok_path)

        ngrok_path_v2_2 = os.path.join(self.config_dir, "v2_2", installer.get_ngrok_bin())
        pyngrok_config_v2_2 = self.copy_with_updates(self.pyngrok_config_v2,
                                                     ngrok_path=ngrok_path_v2_2)
        self.given_ngrok_installed(pyngrok_config_v2_2)
        config_path_v2_2 = os.path.join(self.config_dir, "config_v2_2.yml")
        installer.install_default_config(config_path_v2_2, {"web_addr": "localhost:4041"}, ngrok_version="v2")
        pyngrok_config_v2_2 = self.copy_with_updates(self.pyngrok_config_v2,
                                                     config_path=config_path_v2_2,
                                                     ngrok_path=ngrok_path_v2_2)

        self.given_ngrok_installed(self.pyngrok_config_v3)
        config_path_v3_1 = os.path.join(self.config_dir, "config_v3_1.yml")
        installer.install_default_config(config_path_v3_1, {"web_addr": "localhost:4042"}, ngrok_version="v3")
        pyngrok_config_v3_1 = self.copy_with_updates(self.pyngrok_config_v3,
                                                     config_path=config_path_v3_1,
                                                     ngrok_path=self.pyngrok_config_v3.ngrok_path)

        ngrok_path_v3_2 = os.path.join(self.config_dir, "v3_2", installer.get_ngrok_bin())
        pyngrok_config_v3_2 = self.copy_with_updates(self.pyngrok_config_v3,
                                                     ngrok_path=ngrok_path_v3_2)
        self.given_ngrok_installed(pyngrok_config_v3_2)
        config_path_v3_2 = os.path.join(self.config_dir, "config_v3_2.yml")
        installer.install_default_config(config_path_v3_2, {"web_addr": "localhost:4043"}, ngrok_version="v3")
        pyngrok_config_v3_2 = self.copy_with_updates(self.pyngrok_config_v3,
                                                     config_path=config_path_v3_2,
                                                     ngrok_path=ngrok_path_v3_2)

        # WHEN
        ngrok_process1 = process._start_process(pyngrok_config_v2_1)
        ngrok_process2 = process._start_process(pyngrok_config_v2_2)
        ngrok_process3 = process._start_process(pyngrok_config_v3_1)
        ngrok_process4 = process._start_process(pyngrok_config_v3_2)

        # THEN
        self.assertEqual(len(process._current_processes.keys()), 4)
        self.assertIsNotNone(ngrok_process1)
        self.assertIsNone(ngrok_process1.proc.poll())
        self.assertTrue(ngrok_process1._monitor_thread.is_alive())
        self.assertTrue(urlparse(ngrok_process1.api_url).port, "4040")
        self.assertIsNotNone(ngrok_process2)
        self.assertIsNone(ngrok_process2.proc.poll())
        self.assertTrue(ngrok_process2._monitor_thread.is_alive())
        self.assertTrue(urlparse(ngrok_process2.api_url).port, "4041")
        self.assertIsNotNone(ngrok_process3)
        self.assertIsNone(ngrok_process3.proc.poll())
        self.assertTrue(ngrok_process3._monitor_thread.is_alive())
        self.assertTrue(urlparse(ngrok_process3.api_url).port, "4042")
        self.assertIsNotNone(ngrok_process4)
        self.assertIsNone(ngrok_process4.proc.poll())
        self.assertTrue(ngrok_process4._monitor_thread.is_alive())
        self.assertTrue(urlparse(ngrok_process4.api_url).port, "4043")

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_multiple_processes_same_binary_fails_v2(self):
        # GIVEN
        self.given_ngrok_installed(self.pyngrok_config_v2)
        self.assertEqual(len(process._current_processes.keys()), 0)

        # WHEN
        ngrok_process1 = process._start_process(self.pyngrok_config_v2)
        with self.assertRaises(PyngrokNgrokError) as cm:
            process._start_process(self.pyngrok_config_v2)

        # THEN
        self.assertIn("ngrok is already running", str(cm.exception))
        self.assertIsNotNone(ngrok_process1)
        self.assertIsNone(ngrok_process1.proc.poll())
        self.assertEqual(ngrok_process1, process.get_process(self.pyngrok_config_v2))
        self.assertEqual(len(process._current_processes.keys()), 1)

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_multiple_processes_same_binary_fails_v3(self):
        # GIVEN
        self.given_ngrok_installed(self.pyngrok_config_v3)
        self.assertEqual(len(process._current_processes.keys()), 0)

        # WHEN
        ngrok_process1 = process._start_process(self.pyngrok_config_v3)
        with self.assertRaises(PyngrokNgrokError) as cm:
            process._start_process(self.pyngrok_config_v3)

        # THEN
        self.assertIn("ngrok is already running", str(cm.exception))
        self.assertIsNotNone(ngrok_process1)
        self.assertIsNone(ngrok_process1.proc.poll())
        self.assertEqual(ngrok_process1, process.get_process(self.pyngrok_config_v3))
        self.assertEqual(len(process._current_processes.keys()), 1)

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_process_logs(self):
        # GIVEN
        self.given_ngrok_installed(self.pyngrok_config_v3)

        # WHEN
        ngrok_process = process._start_process(self.pyngrok_config_v3)

        # THEN
        for log in ngrok_process.logs:
            self.assertIsNotNone(log.t)
            self.assertIsNotNone(log.lvl)
            self.assertIsNotNone(log.msg)

    @mock.patch("subprocess.Popen")
    def test_start_new_session(self, mock_popen):
        # GIVEN
        self.given_ngrok_installed(self.pyngrok_config_v3)

        # WHEN
        pyngrok_config = self.copy_with_updates(self.pyngrok_config_v3,
                                                start_new_session=True)
        try:
            process._start_process(pyngrok_config=pyngrok_config)
        except TypeError:
            # Since we're mocking subprocess.Popen, this call will ultimately fail, but it gets far enough for us to
            # validate our assertion
            pass

        # THEN
        self.assertTrue(mock_popen.called)
        if os.name == "posix":
            self.assertIn("start_new_session", mock_popen.call_args[1])
        else:
            self.assertNotIn("start_new_session", mock_popen.call_args[1])

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_log_event_callback_and_max_logs(self):
        # GIVEN
        self.given_ngrok_installed(self.pyngrok_config_v3)
        log_event_callback_mock = mock.MagicMock()
        pyngrok_config = self.copy_with_updates(self.pyngrok_config_v3, log_event_callback=log_event_callback_mock,
                                                max_logs=5)

        # WHEN
        ngrok.connect(pyngrok_config=pyngrok_config)
        ngrok_process = ngrok.get_ngrok_process(pyngrok_config)
        time.sleep(1)

        # THEN
        self.assertGreater(log_event_callback_mock.call_count, len(ngrok_process.logs))
        self.assertEqual(len(ngrok_process.logs), 5)

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_no_monitor_thread(self):
        # GIVEN
        self.given_ngrok_installed(self.pyngrok_config_v3)
        pyngrok_config = self.copy_with_updates(self.pyngrok_config_v3,
                                                monitor_thread=False)

        # WHEN
        ngrok.connect(pyngrok_config=pyngrok_config)
        ngrok_process = ngrok.get_ngrok_process(pyngrok_config)

        # THEN
        self.assertIsNone(ngrok_process._monitor_thread)

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_stop_monitor_thread(self):
        # GIVEN
        self.given_ngrok_installed(self.pyngrok_config_v3)
        current_process = ngrok.get_ngrok_process(pyngrok_config=self.pyngrok_config_v3)
        public_url = ngrok.connect(urlparse(current_process.api_url).port,
                                   pyngrok_config=self.pyngrok_config_v3).public_url
        ngrok_process = ngrok.get_ngrok_process(self.pyngrok_config_v3)
        monitor_thread = ngrok_process._monitor_thread

        # WHEN
        time.sleep(1)
        self.assertTrue(monitor_thread.is_alive())
        ngrok_process.stop_monitor_thread()
        # Make a request to the tunnel to force a log through, which will allow the thread to trigger its own teardown
        urlopen(f"{public_url}/status").read()
        time.sleep(1)

        # THEN
        self.assertFalse(monitor_thread.is_alive())
        self.assertIsNone(ngrok_process._monitor_thread)
        self.assertTrue(ngrok_process.pyngrok_config.monitor_thread)

    @mock.patch("subprocess.Popen")
    def test_retry_session_connection_failure(self, mock_proc_readline):
        # GIVEN
        log_line = "lvl=eror msg=\"some error\" err=EOF"
        mock_proc_readline.return_value.stdout.readline.side_effect = [log_line, log_line, log_line]
        pyngrok_config = self.copy_with_updates(self.pyngrok_config_v3)
        self.given_ngrok_installed(pyngrok_config)

        # WHEN
        with self.assertRaises(PyngrokNgrokError) as cm:
            process._start_process(pyngrok_config)

        # THEN
        self.assertIsNotNone(cm)
        self.assertEqual(cm.exception.ngrok_logs[-1].msg, "some error")
        self.assertEqual(cm.exception.ngrok_error, "EOF")
        self.assertEqual(mock_proc_readline.call_count, 1)
        self.assertEqual(len(process._current_processes.keys()), 0)

    ################################################################################
    # Tests below this point don't need to start a long-lived ngrok process, they
    # are asserting on pyngrok-specific code or edge cases.
    ################################################################################

    def test_start_process_no_binary(self):
        # GIVEN
        self.given_file_doesnt_exist(self.pyngrok_config_v3.ngrok_path)
        self.assertEqual(len(process._current_processes.keys()), 0)

        # WHEN
        with self.assertRaises(PyngrokNgrokError) as cm:
            process.get_process(self.pyngrok_config_v3)

        # THEN
        self.assertIn("ngrok binary was not found", str(cm.exception))
        self.assertEqual(len(process._current_processes.keys()), 0)

    def test_log_parsing(self):
        # GIVEN
        log_line = ("t=2024-03-08T08:45:07-0600 lvl=info msg=\"starting web service\" "
                    "obj=web addr=127.0.0.1:4040 allow_hosts=[]")
        # WHEN
        ngrok_log = NgrokLog(log_line)
        # THEN
        self.assertEqual(ngrok_log.line, log_line)
        self.assertEqual(ngrok_log.t, "2024-03-08T08:45:07-0600")
        self.assertEqual(ngrok_log.lvl, "INFO")
        self.assertEqual(ngrok_log.msg, "starting web service")
        self.assertEqual(ngrok_log.obj, "web")
        self.assertEqual(ngrok_log.addr, "127.0.0.1:4040")

        # WHEN
        ngrok_log = NgrokLog("lvl=INFO msg=Test")
        # THEN
        self.assertEqual(ngrok_log.lvl, "INFO")
        self.assertEqual(ngrok_log.msg, "Test")

        # WHEN
        ngrok_log = NgrokLog("lvl=WARN msg=Test=Test")
        # THEN
        self.assertEqual(ngrok_log.lvl, "WARNING")
        self.assertEqual(ngrok_log.msg, "Test=Test")

        # WHEN
        ngrok_log = NgrokLog("lvl=WARN msg=\"Test=Test with spaces\"")
        # THEN
        self.assertEqual(ngrok_log.lvl, "WARNING")
        self.assertEqual(ngrok_log.msg, "Test=Test with spaces")

        # WHEN
        ngrok_log = NgrokLog("lvl=WARN msg=\"Test=This is Tom's test\"")
        # THEN
        self.assertEqual(ngrok_log.lvl, "WARNING")
        self.assertEqual(ngrok_log.msg, "Test=This is Tom's test")

        # WHEN
        ngrok_log = NgrokLog("lvl=ERR no_msg")
        # THEN
        self.assertEqual(ngrok_log.lvl, "ERROR")
        self.assertIsNone(ngrok_log.msg)

        # WHEN
        ngrok_log = NgrokLog("lvl=CRIT")
        # THEN
        self.assertEqual(ngrok_log.lvl, "CRITICAL")
        self.assertIsNone(ngrok_log.msg)

        # WHEN
        ngrok_log = NgrokLog("lvl=")
        # THEN
        self.assertEqual(ngrok_log.lvl, "NOTSET")

        # WHEN
        ngrok_log = NgrokLog("key=val")
        # THEN
        self.assertEqual(ngrok_log.lvl, "NOTSET")

        # WHEN
        ngrok_log = NgrokLog("lvl=FAKE")
        # THEN
        self.assertEqual(ngrok_log.lvl, "NOTSET")

        # WHEN
        ngrok_log = NgrokLog("t=123456789")
        # THEN
        self.assertEqual(ngrok_log.t, "123456789")



__copyright__ = "Copyright (c) 2018-2025 Alex Laird"
__license__ = "MIT"

import os
import socket
import urllib
import urllib.request
from unittest import mock
from urllib.error import HTTPError

from pyngrok import installer, ngrok, conf
from pyngrok.conf import PyngrokConfig
from pyngrok.exception import PyngrokError, PyngrokNgrokInstallError, PyngrokSecurityError
from pyngrok.installer import PLATFORMS_V3
from tests.testcase import NgrokTestCase


class TestInstaller(NgrokTestCase):
    def test_installer_v2(self):
        # GIVEN
        self.given_file_doesnt_exist(self.pyngrok_config_v2.ngrok_path)
        self.assertFalse(os.path.exists(self.pyngrok_config_v2.ngrok_path))

        # WHEN
        installer.install_ngrok(self.pyngrok_config_v2.ngrok_path, self.pyngrok_config_v2.ngrok_version)
        ngrok_version, pyngrok_version = ngrok.get_version(pyngrok_config=self.pyngrok_config_v2)

        # THEN
        self.assertTrue(os.path.exists(self.pyngrok_config_v2.ngrok_path))
        self.assertTrue(ngrok_version.startswith("2"))

    def test_installer_v3(self):
        # GIVEN
        self.given_file_doesnt_exist(self.pyngrok_config_v3.ngrok_path)
        self.assertFalse(os.path.exists(self.pyngrok_config_v3.ngrok_path))

        # WHEN
        installer.install_ngrok(self.pyngrok_config_v3.ngrok_path, self.pyngrok_config_v3.ngrok_version)
        ngrok_version, pyngrok_version = ngrok.get_version(pyngrok_config=self.pyngrok_config_v3)

        # THEN
        self.assertTrue(os.path.exists(self.pyngrok_config_v3.ngrok_path))
        self.assertTrue(ngrok_version.startswith("3"))

    def test_installer_default(self):
        # GIVEN
        ngrok_path = os.path.join(self.config_dir, "default", installer.get_ngrok_bin())
        config_path = os.path.join(self.config_dir, "config_default.yml")

        pyngrok_config = PyngrokConfig(
            ngrok_path=ngrok_path,
            config_path=config_path)
        self.given_file_doesnt_exist(pyngrok_config.ngrok_path)
        self.given_file_doesnt_exist(pyngrok_config.config_path)
        default_exists = os.path.exists(conf.DEFAULT_NGROK_CONFIG_PATH)
        self.assertFalse(os.path.exists(pyngrok_config.ngrok_path))
        self.assertFalse(os.path.exists(pyngrok_config.config_path))

        # WHEN
        installer.install_ngrok(pyngrok_config.ngrok_path, pyngrok_config.ngrok_version)
        ngrok_version, pyngrok_version = ngrok.get_version(pyngrok_config=pyngrok_config)

        # THEN
        self.assertTrue(os.path.exists(pyngrok_config.ngrok_path))
        self.assertTrue(ngrok_version.startswith("3"))
        self.assertTrue(os.path.exists(pyngrok_config.config_path))
        # Asserting this way ensures CI validates a lack of a default config path, while running
        # locally one a dev's machine (where the default may exist) will not fail the test
        self.assertEqual(default_exists, os.path.exists(conf.DEFAULT_NGROK_CONFIG_PATH))

    def test_config_provisioned(self):
        # GIVEN
        self.given_file_doesnt_exist(self.pyngrok_config_v3.config_path)

        # WHEN
        installer.install_default_config(self.pyngrok_config_v3.config_path, {}, self.pyngrok_config_v3.ngrok_version)

        # THEN
        self.assertTrue(os.path.exists(self.pyngrok_config_v3.config_path))

    def test_get_default_v2_config(self):
        # GIVEN
        installer.install_default_config(self.pyngrok_config_v3.config_path,
                                         {},
                                         self.pyngrok_config_v3.ngrok_version,
                                         "2")

        # WHEN
        ngrok_config = installer.get_ngrok_config(self.pyngrok_config_v3.config_path)

        # THEN
        self.assertEqual(2, len(ngrok_config))
        self.assertEqual("2", ngrok_config["version"])
        self.assertEqual("us", ngrok_config["region"])
        self.assertTrue(os.path.exists(self.pyngrok_config_v3.config_path))

    def test_get_default_v3_config(self):
        # GIVEN
        installer.install_default_config(self.pyngrok_config_v3.config_path,
                                         {},
                                         self.pyngrok_config_v3.ngrok_version,
                                         "3")

        # WHEN
        ngrok_config = installer.get_ngrok_config(self.pyngrok_config_v3.config_path)

        # THEN
        self.assertEqual(1, len(ngrok_config))
        self.assertEqual("3", ngrok_config["version"])
        self.assertTrue(os.path.exists(self.pyngrok_config_v3.config_path))

    ################################################################################
    # Tests below this point don't need to start a long-lived ngrok process, they
    # are asserting on pyngrok-specific code or edge cases.
    ################################################################################

    @mock.patch("pyngrok.installer.urlopen")
    def test_installer_download_fails(self, mock_urlopen):
        # GIVEN
        magic_mock = mock.MagicMock()
        magic_mock.getcode.return_value = 500
        mock_urlopen.return_value = magic_mock

        self.given_file_doesnt_exist(self.pyngrok_config_v3.ngrok_path)
        self.assertFalse(os.path.exists(self.pyngrok_config_v3.ngrok_path))

        # WHEN
        with self.assertRaises(PyngrokNgrokInstallError):
            ngrok.connect(pyngrok_config=self.pyngrok_config_v3)

        # THEN
        self.assertFalse(os.path.exists(self.pyngrok_config_v3.ngrok_path))

    @mock.patch("pyngrok.installer.urlopen")
    def test_installer_retry(self, mock_urlopen):
        # GIVEN
        mock_urlopen.side_effect = socket.timeout("The read operation timed out")

        self.given_file_doesnt_exist(self.pyngrok_config_v3.ngrok_path)
        self.assertFalse(os.path.exists(self.pyngrok_config_v3.ngrok_path))

        # WHEN
        with self.assertRaises(PyngrokNgrokInstallError):
            ngrok.connect(pyngrok_config=self.pyngrok_config_v3)

        # THEN
        self.assertEqual(mock_urlopen.call_count, 4)
        self.assertFalse(os.path.exists(self.pyngrok_config_v3.ngrok_path))

    def test_download_file_security_error(self):
        # WHEN
        with self.assertRaises(PyngrokSecurityError):
            installer._download_file(f"file:{__file__}", retries=10)

    def test_web_addr_false_not_allowed(self):
        # WHEN
        with self.assertRaises(PyngrokError):
            installer.install_default_config(self.pyngrok_config_v3.config_path, {"web_addr": False})

    @mock.patch("platform.system")
    def test_get_default_ngrok_dir_mac(self, mock_system):
        # GIVEN
        mock_system.return_value = "Darwin"
        user_home = os.path.expanduser("~")

        # WHEN
        default_ngrok_dir = installer.get_default_ngrok_dir()

        # THEN
        self.assertEqual(os.path.join(user_home, "Library", "Application Support", "ngrok"), default_ngrok_dir)

    @mock.patch("platform.system")
    def test_get_default_ngrok_dir_windows(self, mock_system):
        # GIVEN
        mock_system.return_value = "Windows 10"
        user_home = os.path.expanduser("~")

        # WHEN
        default_ngrok_dir = installer.get_default_ngrok_dir()

        # THEN
        self.assertEqual(os.path.join(user_home, "AppData", "Local", "ngrok"), default_ngrok_dir)

    @mock.patch("platform.system")
    def test_get_default_ngrok_dir_unix(self, mock_system):
        # GIVEN
        mock_system.return_value = "Linux"
        user_home = os.path.expanduser("~")

        # WHEN
        default_ngrok_dir = installer.get_default_ngrok_dir()

        # THEN
        self.assertEqual(os.path.join(user_home, ".config", "ngrok"), default_ngrok_dir)

    @mock.patch("platform.system")
    def test_get_ngrok_binary_mac(self, mock_system):
        # GIVEN
        mock_system.return_value = "Darwin"

        # WHEN
        ngrok_bin = installer.get_ngrok_bin()

        # THEN
        self.assertEqual("ngrok", ngrok_bin)

    @mock.patch("platform.system")
    def test_get_system_windows(self, mock_system):
        # GIVEN
        mock_system.return_value = "Windows 10"

        # WHEN
        ngrok_bin = installer.get_ngrok_bin()

        # THEN
        self.assertEqual("ngrok.exe", ngrok_bin)

    @mock.patch("platform.system")
    def test_get_system_linux(self, mock_system):
        # GIVEN
        mock_system.return_value = "Linux"

        # WHEN
        ngrok_bin = installer.get_ngrok_bin()

        # THEN
        self.assertEqual("ngrok", ngrok_bin)

    @mock.patch("platform.system")
    def test_get_system_freebsd(self, mock_system):
        # GIVEN
        mock_system.return_value = "FreeBSD"

        # WHEN
        ngrok_bin = installer.get_ngrok_bin()

        # THEN
        self.assertEqual("ngrok", ngrok_bin)

    @mock.patch("platform.system")
    def test_get_system_cygwin_is_windows(self, mock_system):
        # GIVEN
        mock_system.return_value = "Cygwin NT"

        # WHEN
        ngrok_bin = installer.get_ngrok_bin()

        # THEN
        self.assertEqual("ngrok.exe", ngrok_bin)

    @mock.patch("platform.system")
    def test_get_system_mingw_is_windows(self, mock_system):
        # GIVEN
        mock_system.return_value = "MinGW NT"

        # WHEN
        ngrok_bin = installer.get_ngrok_bin()

        # THEN
        self.assertEqual("ngrok.exe", ngrok_bin)

    @mock.patch("platform.system")
    def test_get_system_unsupported(self, mock_system):
        # GIVEN
        mock_system.return_value = "Solaris"

        # WHEN
        with self.assertRaises(PyngrokNgrokInstallError):
            installer.get_ngrok_bin()

    @mock.patch("platform.system")
    @mock.patch("platform.machine")
    def test_get_ngrok_cdn_url_mac_arm64(self, mock_machine, mock_system):
        # GIVEN
        mock_system.return_value = "Darwin"
        mock_machine.return_value = "arm64"

        # WHEN
        url = installer.get_ngrok_cdn_url("v3")

        # THEN
        self.assertEqual(PLATFORMS_V3["darwin_x86_64_arm"], url)

    @mock.patch("platform.system")
    @mock.patch("platform.machine")
    def test_get_ngrok_cdn_url_windows32(self, mock_machine, mock_system):
        # GIVEN
        mock_system.return_value = "Windows 10"
        mock_machine.return_value = "i386"

        # WHEN
        url = installer.get_ngrok_cdn_url("v3")

        # THEN
        self.assertEqual(PLATFORMS_V3["windows_i386"], url)

    @mock.patch("platform.system")
    @mock.patch("platform.machine")
    def test_get_ngrok_cdn_url_linux_arm64(self, mock_machine, mock_system):
        # GIVEN
        mock_system.return_value = "Linux"
        mock_machine.return_value = "arm aarch64"

        # WHEN
        url = installer.get_ngrok_cdn_url("v3")

        # THEN
        self.assertEqual(PLATFORMS_V3["linux_x86_64_arm"], url)

    @mock.patch("platform.system")
    @mock.patch("platform.machine")
    def test_get_ngrok_cdn_url_freebsd64(self, mock_machine, mock_system):
        # GIVEN
        mock_system.return_value = "FreeBSD"
        mock_machine.return_value = "x86_64"

        # WHEN
        url = installer.get_ngrok_cdn_url("v3")

        # THEN
        self.assertEqual(PLATFORMS_V3["freebsd_x86_64"], url)

    @mock.patch("platform.system")
    @mock.patch("platform.machine")
    def test_get_ngrok_cdn_url_cygwin_is_windows(self, mock_machine, mock_system):
        # GIVEN
        mock_system.return_value = "Cygwin NT"
        mock_machine.return_value = "aarch64"

        # WHEN
        url = installer.get_ngrok_cdn_url("v3")

        # THEN
        self.assertEqual(PLATFORMS_V3["windows_x86_64_arm"], url)

    @mock.patch("platform.system")
    @mock.patch("platform.machine")
    def test_get_ngrok_cdn_url_mingw_is_windows(self, mock_machine, mock_system):
        # GIVEN
        mock_system.return_value = "MinGW NT"
        mock_machine.return_value = "i386"

        # WHEN
        url = installer.get_ngrok_cdn_url("v3")

        # THEN
        self.assertEqual(PLATFORMS_V3["windows_i386"], url)

    def test_ensure_installer_urls_exist_v2(self):
        for key, value in installer.PLATFORMS.items():
            try:
                urllib.request.urlopen(urllib.request.Request(value, method="HEAD"))
            except HTTPError as e:
                if e.status == 404:
                    self.fail("Installer URL returned 404: %s" % value)
                else:
                    raise e

    def test_ensure_installer_urls_exist_v3(self):
        for key, value in installer.PLATFORMS_V3.items():
            try:
                urllib.request.urlopen(urllib.request.Request(value, method="HEAD"))
            except HTTPError as e:
                if e.status == 404:
                    self.fail("Installer URL returned 404: %s" % value)
                else:
                    raise e



#!/usr/bin/env python

__copyright__ = "Copyright (c) 2018-2025 Alex Laird"
__license__ = "MIT"

import json
import logging
import os
import socket
import sys
import uuid
from http import HTTPStatus
from typing import Any, Dict, List, Optional, Tuple, Union
from urllib.error import HTTPError, URLError
from urllib.parse import urlencode
from urllib.request import Request, urlopen

from pyngrok import __version__, conf, installer, process
from pyngrok.conf import PyngrokConfig
from pyngrok.exception import PyngrokError, PyngrokNgrokHTTPError, PyngrokNgrokURLError, PyngrokSecurityError
from pyngrok.installer import get_default_config
from pyngrok.process import NgrokProcess

logger = logging.getLogger(__name__)


class NgrokTunnel:
    """
    An object containing information about a ``ngrok`` tunnel.
    """

    def __init__(self,
                 data: Dict[str, Any],
                 pyngrok_config: PyngrokConfig,
                 api_url: Optional[str]) -> None:
        #: The original tunnel data.
        self.data: Dict[str, Any] = data
        #: The ``pyngrok`` configuration to use when interacting with the ``ngrok``.
        self.pyngrok_config: PyngrokConfig = pyngrok_config
        #: The API URL for the ``ngrok`` web interface.
        self.api_url: Optional[str] = api_url

        #: The ID of the tunnel.
        self.id: Optional[str] = data.get("ID", None)
        #: The name of the tunnel.
        self.name: Optional[str] = data.get("name")
        #: The protocol of the tunnel.
        self.proto: Optional[str] = data.get("proto")
        #: The tunnel URI, a relative path that can be used to make requests to the ``ngrok`` web interface.
        self.uri: Optional[str] = data.get("uri")
        #: The public ``ngrok`` URL.
        self.public_url: Optional[str] = data.get("public_url")
        #: The config for the tunnel.
        self.config: Dict[str, Any] = data.get("config", {})
        #: Metrics for `the tunnel <https://ngrok.com/docs/agent/api/#list-tunnels>`_.
        self.metrics: Dict[str, Any] = data.get("metrics", {})

    def __repr__(self) -> str:
        return f"<NgrokTunnel: \"{self.public_url}\" -> \"{self.config['addr']}\">" if self.config.get(
            "addr", None) else "<pending Tunnel>"

    def __str__(self) -> str:  # pragma: no cover
        return f"NgrokTunnel: \"{self.public_url}\" -> \"{self.config['addr']}\"" if self.config.get(
            "addr", None) else "<pending Tunnel>"

    def refresh_metrics(self) -> None:
        """
        Get the latest metrics for the tunnel and update the ``metrics`` variable.

        :raises: :class:`~pyngrok.exception.PyngrokError`: When the API does not return ``metrics``.
        """
        logger.info(f"Refreshing metrics for tunnel: {self.public_url}")

        data = api_request(f"{self.api_url}{self.uri}", method="GET",
                           timeout=self.pyngrok_config.request_timeout)

        if "metrics" not in data:
            raise PyngrokError("The ngrok API did not return \"metrics\" in the response")

        self.data["metrics"] = data["metrics"]
        self.metrics = self.data["metrics"]


class NgrokApiResponse:
    """
    An object containing a response from the ``ngrok`` API.
    """

    def __init__(self,
                 status: str,
                 data: Optional[Dict[str, Any]]) -> None:
        #: The description of the response.
        self.status: str = status
        #: The parsed API response.
        self.data: Optional[Dict[str, Any]] = data

    @staticmethod
    def from_body(body: str) -> "NgrokApiResponse":
        """
        Construct an object from a response body.

        :param body: The response body to be parsed.
        :return: The constructed object.
        """
        json_starts = body.find("{")

        if json_starts < 0:
            return NgrokApiResponse(body, None)
        else:
            return NgrokApiResponse(body[:json_starts], json.loads(body[json_starts:]))


_current_tunnels: Dict[str, NgrokTunnel] = {}


def install_ngrok(pyngrok_config: Optional[PyngrokConfig] = None) -> None:
    """
    Download, install, and initialize ``ngrok`` for the given config. If ``ngrok`` and its default
    config is already installed, calling this method will do nothing.

    :param pyngrok_config: A ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
        overriding :func:`~pyngrok.conf.get_default()`.
    """
    if pyngrok_config is None:
        pyngrok_config = conf.get_default()

    if not os.path.exists(pyngrok_config.ngrok_path):
        installer.install_ngrok(pyngrok_config.ngrok_path, ngrok_version=pyngrok_config.ngrok_version)

    config_path = conf.get_config_path(pyngrok_config)

    # Install the config to the requested path
    with installer.config_file_lock:
        if not os.path.exists(config_path):
            installer.install_default_config(config_path, ngrok_version=pyngrok_config.ngrok_version)


def set_auth_token(token: str,
                   pyngrok_config: Optional[PyngrokConfig] = None) -> None:
    """
    Set the ``ngrok`` auth token in the config file to streamline access to more features (for instance, multiple
    concurrent tunnels, custom domains, etc.).

    If ``ngrok`` is not installed at :class:`~pyngrok.conf.PyngrokConfig`'s ``ngrok_path``, calling this method
    will first download and install ``ngrok``.

    :param token: The auth token to set.
    :param pyngrok_config: A ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
        overriding :func:`~pyngrok.conf.get_default()`.
    """
    if pyngrok_config is None:
        pyngrok_config = conf.get_default()

    install_ngrok(pyngrok_config)

    process.set_auth_token(pyngrok_config, token)


def set_api_key(key: str,
                pyngrok_config: Optional[PyngrokConfig] = None) -> None:
    """
    Set the ``ngrok`` API key in the config file to enable access to more features (for instance,
    `Internal Endpoints <https://ngrok.com/docs/universal-gateway/internal-endpoints/>`_).

    If ``ngrok`` is not installed at :class:`~pyngrok.conf.PyngrokConfig`'s ``ngrok_path``, calling this method
    will first download and install ``ngrok``.

    :param key: The API key to set.
    :param pyngrok_config: A ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
        overriding :func:`~pyngrok.conf.get_default()`.
    """
    if pyngrok_config is None:
        pyngrok_config = conf.get_default()

    install_ngrok(pyngrok_config)

    process.set_api_key(pyngrok_config, key)


def get_ngrok_process(pyngrok_config: Optional[PyngrokConfig] = None) -> NgrokProcess:
    """
    Get the current ``ngrok`` process for the given config's ``ngrok_path``.

    If ``ngrok`` is not installed at :class:`~pyngrok.conf.PyngrokConfig`'s ``ngrok_path``, calling this method
    will first download and install ``ngrok``.

    If ``ngrok`` is not running, calling this method will first start a process with
    :class:`~pyngrok.conf.PyngrokConfig`.

    Use :func:`~pyngrok.process.is_process_running` to check if a process is running without also implicitly
    installing and starting it.

    :param pyngrok_config: A ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
        overriding :func:`~pyngrok.conf.get_default()`.
    :return: The ``ngrok`` process.
    """
    if pyngrok_config is None:
        pyngrok_config = conf.get_default()

    install_ngrok(pyngrok_config)

    return process.get_process(pyngrok_config)


def _interpolate_tunnel_definition(pyngrok_config: PyngrokConfig,
                                   options: Dict[str, Any],
                                   addr: Optional[str] = None,
                                   proto: Optional[Union[str, int]] = None,
                                   name: Optional[str] = None) -> None:
    config_path = conf.get_config_path(pyngrok_config)

    with installer.config_file_lock:
        if os.path.exists(config_path):
            config = installer.get_ngrok_config(config_path, ngrok_version=pyngrok_config.ngrok_version)
        else:
            config = get_default_config(pyngrok_config.ngrok_version, pyngrok_config.config_version)

    tunnel_definitions = config.get("tunnels", {})
    # If a "pyngrok-default" tunnel definition exists in the ngrok config, use that
    if not name and "pyngrok-default" in tunnel_definitions:
        logger.info("pyngrok-default found defined in config, using for tunnel definition")
        name = "pyngrok-default"

    # Use a tunnel definition for the given name, if it exists
    if name and name in tunnel_definitions:
        tunnel_definition = tunnel_definitions[name]

        addr = tunnel_definition.get("addr") if not addr else addr
        proto = tunnel_definition.get("proto") if not proto else proto
        # Use the tunnel definition as the base, but override with any passed in options
        tunnel_definition.update(options)
        options.clear()
        options.update(tunnel_definition)
        name += "-api"

    addr = str(addr) if addr else "80"
    if not proto:
        proto = "http"

    if not name:
        if not addr.startswith("file://"):
            name = f"{proto}-{addr}-{uuid.uuid4()}"
        else:
            name = f"{proto}-file-{uuid.uuid4()}"

    options.update({
        "name": name,
        "addr": addr,
        "proto": proto
    })


def _upgrade_legacy_params(pyngrok_config: PyngrokConfig,
                           options: Dict[str, Any]) -> None:
    if pyngrok_config.ngrok_version == "v3":
        if "bind_tls" in options:
            if options.get("bind_tls") is True or options.get("bind_tls") == "true":
                options["schemes"] = ["https"]
            elif not options.get("bind_tls") is not False or options.get("bind_tls") == "false":
                options["schemes"] = ["http"]
            else:
                options["schemes"] = ["http", "https"]

            options.pop("bind_tls")

        if "auth" in options:
            auth = options.get("auth")
            if isinstance(auth, list):
                options["basic_auth"] = auth
            else:
                options["basic_auth"] = [auth]

            options.pop("auth")


def connect(addr: Optional[str] = None,
            proto: Optional[Union[str, int]] = None,
            name: Optional[str] = None,
            pyngrok_config: Optional[PyngrokConfig] = None,
            **options: Any) -> NgrokTunnel:
    """
    Establish a new ``ngrok`` tunnel for the given protocol to the given port, returning an object representing
    the connected tunnel.

    If a `tunnel definition in ngrok's config file
    <https://ngrok.com/docs/agent/config/v2/#tunnel-configurations>`_ matches the given
    ``name``, it will be loaded and used to start the tunnel (note that "-api" will be appended to its name when
    started). When ``name`` is ``None`` and a "pyngrok-default" tunnel definition exists in ``ngrok``'s config, it
    will be loaded and used. Any ``kwargs`` passed as ``options`` will override properties from the loaded
    tunnel definition.

    If ``ngrok`` is not installed at :class:`~pyngrok.conf.PyngrokConfig`'s ``ngrok_path``, calling this method
    will first download and install ``ngrok``.

    ``pyngrok`` is compatible with ``ngrok`` v2 and v3, but by default it will install v3. To install v2 instead,
    set ``ngrok_version`` to "v2" in :class:`~pyngrok.conf.PyngrokConfig`:

    If ``ngrok`` is not running, calling this method will first start a process with
    :class:`~pyngrok.conf.PyngrokConfig`.

    .. note::

        ``ngrok`` v2's default behavior for ``http`` when no additional properties are passed is to open *two* tunnels,
        one ``http`` and one ``https``. This method will return a reference to the ``http`` tunnel in this case. If
        only a single tunnel is needed, pass ``bind_tls=True`` and a reference to the ``https`` tunnel will be
        returned.

    :param addr: The local port to which the tunnel will forward traffic, or a
        `local directory or network address <https://ngrok.com/docs/http/#file-serving>`_,
        defaults to "80".
    :param proto: A valid `tunnel protocol
        <https://ngrok.com/docs/agent/config/v2/#tunnel-configurations>`_, defaults to "http".
    :param name: A friendly name for the tunnel, or the name of a `ngrok tunnel definition
        <https://ngrok.com/docs/agent/config/v2/#tunnel-configurations>`_ defined in ``ngrok``'s config file.
    :param pyngrok_config: A ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
        overriding :func:`~pyngrok.conf.get_default()`.
    :param options: Remaining ``kwargs`` are passed as `configuration for the ngrok
        tunnel <https://ngrok.com/docs/agent/config/v2/#tunnel-configurations>`_.
    :return: The created ``ngrok`` tunnel.
    :raises: :class:`~pyngrok.exception.PyngrokError`: When the tunnel definition is invalid, or the response does
        not contain ``public_url``.
    """
    if pyngrok_config is None:
        pyngrok_config = conf.get_default()

    _interpolate_tunnel_definition(pyngrok_config, options, addr, proto, name)

    proto = options.get("proto")
    name = options.get("name")

    _upgrade_legacy_params(pyngrok_config, options)

    logger.info(f"Opening tunnel named: {name}")

    api_url = get_ngrok_process(pyngrok_config).api_url

    logger.debug(f"Creating tunnel with options: {options}")

    tunnel = NgrokTunnel(api_request(f"{api_url}/api/tunnels", method="POST", data=options,
                                     timeout=pyngrok_config.request_timeout),
                         pyngrok_config, api_url)

    if pyngrok_config.ngrok_version == "v2" and proto == "http" and options.get("bind_tls", "both") == "both":
        tunnel = NgrokTunnel(api_request(f"{api_url}{tunnel.uri}%20%28http%29", method="GET",
                                         timeout=pyngrok_config.request_timeout),
                             pyngrok_config, api_url)

        logger.info(f"ngrok v2 opens multiple tunnels, fetching just HTTP tunnel {tunnel.id} for return")

    if tunnel.public_url is None:
        raise PyngrokError(
            f"\"public_url\" was not populated for tunnel {tunnel}, but is required for pyngrok to function.")

    _current_tunnels[tunnel.public_url] = tunnel

    return tunnel


def disconnect(public_url: str,
               pyngrok_config: Optional[PyngrokConfig] = None) -> None:
    """
    Disconnect the ``ngrok`` tunnel for the given URL, if open.

    If ``ngrok`` is not running, calling this method will first start a process with
    :class:`~pyngrok.conf.PyngrokConfig`.

    :param public_url: The public URL of the tunnel to disconnect.
    :param pyngrok_config: A ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
        overriding :func:`~pyngrok.conf.get_default()`.
    """
    if pyngrok_config is None:
        pyngrok_config = conf.get_default()

    # If ngrok is not running, there are no tunnels to disconnect
    if not process.is_process_running(pyngrok_config.ngrok_path):
        logger.debug(f"\"ngrok_path\" {pyngrok_config.ngrok_path} is not running a process")

        return

    api_url = get_ngrok_process(pyngrok_config).api_url

    if public_url not in _current_tunnels:
        get_tunnels(pyngrok_config)

        # One more check, if the given URL is still not in the list of tunnels, it is not active
        if public_url not in _current_tunnels:
            return

    tunnel = _current_tunnels[public_url]

    logger.info(f"Disconnecting tunnel: {tunnel.public_url}")

    api_request(f"{api_url}{tunnel.uri}", method="DELETE",
                timeout=pyngrok_config.request_timeout)

    _current_tunnels.pop(public_url, None)


def get_tunnels(pyngrok_config: Optional[PyngrokConfig] = None) -> List[NgrokTunnel]:
    """
    Get a list of active ``ngrok`` tunnels for the given config's ``ngrok_path``.

    If ``ngrok`` is not installed at :class:`~pyngrok.conf.PyngrokConfig`'s ``ngrok_path``, calling this method
    will first download and install ``ngrok``.

    If ``ngrok`` is not running, calling this method will first start a process with
    :class:`~pyngrok.conf.PyngrokConfig`.

    :param pyngrok_config: A ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
        overriding :func:`~pyngrok.conf.get_default()`.
    :return: The active ``ngrok`` tunnels.
    :raises: :class:`~pyngrok.exception.PyngrokError`: When the response was invalid or does not
        contain ``public_url``.
    """
    if pyngrok_config is None:
        pyngrok_config = conf.get_default()

    api_url = get_ngrok_process(pyngrok_config).api_url

    _current_tunnels.clear()
    for tunnel in api_request(f"{api_url}/api/tunnels", method="GET",
                              timeout=pyngrok_config.request_timeout)["tunnels"]:
        ngrok_tunnel = NgrokTunnel(tunnel, pyngrok_config, api_url)

        if ngrok_tunnel.public_url is None:
            raise PyngrokError(
                f"\"public_url\" was not populated for tunnel {ngrok_tunnel}, "
                f"but is required for pyngrok to function.")

        _current_tunnels[ngrok_tunnel.public_url] = ngrok_tunnel

    return list(_current_tunnels.values())


def kill(pyngrok_config: Optional[PyngrokConfig] = None) -> None:
    """
    Terminate the ``ngrok`` processes, if running, for the given config's ``ngrok_path``. This method will not
    block, it will just issue a kill request.

    :param pyngrok_config: A ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
        overriding :func:`~pyngrok.conf.get_default()`.
    """
    if pyngrok_config is None:
        pyngrok_config = conf.get_default()

    process.kill_process(pyngrok_config.ngrok_path)

    _current_tunnels.clear()


def api(*args: Any, pyngrok_config: Optional[PyngrokConfig] = None) -> NgrokApiResponse:
    """
    Run a ``ngrok`` command against the ``api`` with the given args. This will use the local agent
    to run a remote API request for ``ngrok``, which requires that an API key has been set. For a list of
    available commands, pass ``"--help"``.

    :param pyngrok_config: A ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
        overriding :func:`~pyngrok.conf.get_default()`.
    :param args: The args to pass to the ``api`` command.
    :return: The response from executing the ``api`` command.
    :raises: PyngrokNgrokError The ``ngrok`` process exited with an error.
    :raises: CalledProcessError An error occurred while executing the process.
    """
    if pyngrok_config is None:
        pyngrok_config = conf.get_default()

    cmd_args = ["--config", pyngrok_config.config_path] if pyngrok_config.config_path else []
    cmd_args.append("api")
    if pyngrok_config.api_key:
        cmd_args += ["--api-key", pyngrok_config.api_key]
    cmd_args += [*args]

    logger.info(f"Executing \"ngrok api\" command with args: {args}")

    return NgrokApiResponse.from_body(
        process.capture_run_process(pyngrok_config.ngrok_path,
                                    cmd_args))


def get_version(pyngrok_config: Optional[PyngrokConfig] = None) -> Tuple[str, str]:
    """
    Get a tuple with the ``ngrok`` and ``pyngrok`` versions.

    :param pyngrok_config: A ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
        overriding :func:`~pyngrok.conf.get_default()`.
    :return: A tuple of ``(ngrok_version, pyngrok_version)``.
    """
    if pyngrok_config is None:
        pyngrok_config = conf.get_default()

    install_ngrok(pyngrok_config)

    ngrok_version = process.capture_run_process(pyngrok_config.ngrok_path, ["--version"]).split("version ")[1]

    return ngrok_version, __version__


def update(pyngrok_config: Optional[PyngrokConfig] = None) -> str:
    """
    Update ``ngrok`` for the given config's ``ngrok_path``, if an update is available.

    :param pyngrok_config: A ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
        overriding :func:`~pyngrok.conf.get_default()`.
    :return: The result from the ``ngrok`` update.
    """
    if pyngrok_config is None:
        pyngrok_config = conf.get_default()

    install_ngrok(pyngrok_config)

    return process.capture_run_process(pyngrok_config.ngrok_path, ["update"])


def api_request(url: str,
                method: str = "GET",
                data: Optional[Dict[str, Any]] = None,
                params: Optional[Dict[str, Any]] = None,
                timeout: float = 4,
                auth: Optional[str] = None) -> Dict[str, Any]:
    """
    Invoke an API request to the given URL, returning JSON data from the response.

    One use for this method is making requests to ``ngrok`` tunnels:

    .. code-block:: python

        from pyngrok import ngrok

        public_url = ngrok.connect()
        response = ngrok.api_request(f"{public_url}/some-route",
                                     method="POST", data={"foo": "bar"})

    Another is making requests to the ``ngrok`` API itself:

    .. code-block:: python

        from pyngrok import ngrok

        api_url = ngrok.get_ngrok_process().api_url
        response = ngrok.api_request(f"{api_url}/api/requests/http",
                                     params={"tunnel_name": "foo"})

    :param url: The request URL.
    :param method: The HTTP method.
    :param data: The request body.
    :param params: The URL parameters.
    :param timeout: The request timeout, in seconds.
    :param auth: Set as Bearer for an Authorization header.
    :return: The response from the request.
    :raises: :class:`~pyngrok.exception.PyngrokSecurityError`: When the ``url`` is not supported.
    :raises: :class:`~pyngrok.exception.PyngrokNgrokHTTPError`: When the request returns an error response.
    :raises: :class:`~pyngrok.exception.PyngrokNgrokURLError`: When the request times out.
    """
    if params is None:
        params = {}

    if not url.lower().startswith("http"):
        raise PyngrokSecurityError(f"URL must start with \"http\": {url}")

    encoded_data = json.dumps(data).encode("utf-8") if data else None

    if params:
        url += f"?{urlencode([(x, params[x]) for x in params])}"

    request = Request(url, method=method.upper())
    request.add_header("Content-Type", "application/json")
    if auth:
        request.add_header("Ngrok-Version", "2")
        request.add_header("Authorization", f"Bearer {auth}")

    logger.debug(f"Making {method} request to {url} with data: {data}")

    try:
        response = urlopen(request, encoded_data, timeout)
        response_data = response.read().decode("utf-8")

        status_code = response.getcode()
        logger.debug(f"Response {status_code}: {response_data.strip()}")

        if str(status_code)[0] != "2":
            raise PyngrokNgrokHTTPError(f"ngrok client API returned {status_code}: {response_data}", url,
                                        status_code, None, request.headers, response_data)
        elif status_code == HTTPStatus.NO_CONTENT:
            return {}

        return json.loads(response_data)  # type: ignore
    except socket.timeout:
        raise PyngrokNgrokURLError("ngrok client exception, URLError: timed out", "timed out")
    except HTTPError as e:
        response_data = e.read().decode("utf-8")

        status_code = e.getcode()
        logger.debug(f"Response {status_code}: {response_data.strip()}")

        raise PyngrokNgrokHTTPError(f"ngrok client exception, API returned {status_code}: {response_data}",
                                    e.url,
                                    status_code, e.reason, e.headers, response_data)
    except URLError as e:
        raise PyngrokNgrokURLError(f"ngrok client exception, URLError: {e.reason}", e.reason)


def run(args: Optional[List[str]] = None,
        pyngrok_config: Optional[PyngrokConfig] = None) -> None:
    """
    Ensure ``ngrok`` is installed at the default path, then call :func:`~pyngrok.process.run_process`.

    This method is meant for interacting with ``ngrok`` from the command line and is not necessarily
    compatible with non-blocking API methods. For that, use :mod:`~pyngrok.ngrok`'s interface methods (like
    :func:`~pyngrok.ngrok.connect`), or use :func:`~pyngrok.process.get_process`.

    :param args: Arguments to be passed to the ``ngrok`` process.
    :param pyngrok_config: A ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
        overriding :func:`~pyngrok.conf.get_default()`.
    """
    if args is None:
        args = []
    if pyngrok_config is None:
        pyngrok_config = conf.get_default()

    install_ngrok(pyngrok_config)

    process.run_process(pyngrok_config.ngrok_path, args)


def main() -> None:
    """
    Entry point for the package's ``console_scripts``. This initializes a call from the command
    line and invokes :func:`~pyngrok.ngrok.run`.

    This method is meant for interacting with ``ngrok`` from the command line and is not necessarily
    compatible with non-blocking API methods. For that, use :mod:`~pyngrok.ngrok`'s interface methods (like
    :func:`~pyngrok.ngrok.connect`), or use :func:`~pyngrok.process.get_process`.
    """
    run(sys.argv[1:])

    if len(sys.argv) == 1 or len(sys.argv) == 2 and sys.argv[1].lstrip("-").lstrip("-") == "help":
        print(f"\nPYNGROK VERSION:\n   {__version__}")
    elif len(sys.argv) == 2 and sys.argv[1].lstrip("-").lstrip("-") in ["v", "version"]:
        print(f"pyngrok version {__version__}")


if __name__ == "__main__":
    main()



__copyright__ = "Copyright (c) 2018-2024 Alex Laird"
__license__ = "MIT"

import logging
import os
import shutil
import unittest
from copy import copy

import psutil
from psutil import AccessDenied, NoSuchProcess

from pyngrok import conf, installer, ngrok, process
from pyngrok.conf import PyngrokConfig

logger = logging.getLogger(__name__)
ngrok_logger = logging.getLogger(f"{__name__}.ngrok")


class NgrokTestCase(unittest.TestCase):
    def setUp(self):
        self.config_dir = os.path.normpath(os.path.join(os.path.abspath(os.path.dirname(__file__)), ".ngrok"))
        if not os.path.exists(self.config_dir):
            os.makedirs(self.config_dir)
        config_v2_path = os.path.join(self.config_dir, "config_v2.yml")

        conf.DEFAULT_NGROK_CONFIG_PATH = config_v2_path
        conf.DEFAULT_NGROK_PATH = os.path.join(config_v2_path, installer.get_ngrok_bin())

        config_v3_path = os.path.join(self.config_dir, "config_v3.yml")

        ngrok_path_v2 = os.path.join(self.config_dir, "v2", installer.get_ngrok_bin())
        self.pyngrok_config_v2 = PyngrokConfig(
            ngrok_path=ngrok_path_v2,
            config_path=config_v2_path,
            ngrok_version="v2")

        ngrok_path_v3 = os.path.join(self.config_dir, "v3", installer.get_ngrok_bin())
        self.pyngrok_config_v3 = PyngrokConfig(ngrok_path=ngrok_path_v3,
                                               config_path=config_v3_path,
                                               ngrok_version="v3")

        conf.set_default(self.pyngrok_config_v2)

        # ngrok's CDN can be flaky, so make sure its flakiness isn't reflect in our CI/CD test runs
        installer.DEFAULT_RETRY_COUNT = 3

    def tearDown(self):
        for p in list(process._current_processes.values()):
            try:
                process.kill_process(p.pyngrok_config.ngrok_path)
                p.proc.wait()
            except OSError:
                pass

        ngrok._current_tunnels.clear()

        if os.path.exists(self.config_dir):
            shutil.rmtree(self.config_dir)

    @staticmethod
    def given_ngrok_installed(pyngrok_config):
        ngrok.install_ngrok(pyngrok_config)

    @staticmethod
    def given_file_doesnt_exist(path):
        if os.path.exists(path):
            os.remove(path)

    @staticmethod
    def copy_with_updates(to_copy, **kwargs):
        copied = copy(to_copy)

        for key, value in kwargs.items():
            copied.__setattr__(key, value)

        return copied

    def assert_no_zombies(self):
        try:
            self.assertEqual(0, len(
                list(filter(lambda p: p.name() == "ngrok" and p.status() == "zombie", psutil.process_iter()))))
        except (AccessDenied, NoSuchProcess):
            # Some OSes are flaky on this assertion, but that isn't an indication anything is wrong, so pass
            pass
